# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Portions Copyright Buildbot Team Members
# Portions Copyright 2014 Longaccess private company

import os

from twisted.trial import unittest

from buildbot.test.util.warnings import assertNotProducesWarnings
from buildbot.warnings import DeprecatedApiWarning

try:
    import boto3
    from botocore.client import ClientError
    from moto import mock_aws
except ImportError:
    boto3 = None
    ClientError = None


if boto3 is not None:
    from buildbot.worker import ec2
else:
    ec2 = None  # type: ignore[assignment]


# Current moto (1.3.7) requires dummy credentials to work
# https://github.com/spulec/moto/issues/1924
os.environ['AWS_SECRET_ACCESS_KEY'] = 'foobar_secret'
os.environ['AWS_ACCESS_KEY_ID'] = 'foobar_key'
os.environ['AWS_DEFAULT_REGION'] = 'us-east-1'


# redefine the mock_aws decorator to skip the test if boto3 or moto
# isn't installed
def skip_ec2(f):
    f.skip = "boto3 or moto is not installed"
    return f


if boto3 is None:
    mock_aws = skip_ec2  # type: ignore[assignment]


def anyImageId(c):
    for image in c.describe_images()['Images']:
        return image['ImageId']
    return 'foo'


class TestEC2LatentWorker(unittest.TestCase):
    ec2_connection = None

    def setUp(self):
        super().setUp()
        if boto3 is None:
            raise unittest.SkipTest("moto not found")

    def botoSetup(self, name='latent_buildbot_worker'):
        # the proxy system is also not properly mocked, so we need to delete environment variables
        for env in ['http_proxy', 'https_proxy', 'HTTP_PROXY', 'HTTPS_PROXY']:
            if env in os.environ:
                del os.environ[env]
        # create key pair is not correctly mocked and need to have fake aws creds configured
        kw = {
            "region_name": 'us-east-1',
            "aws_access_key_id": 'ACCESS_KEY',
            "aws_secret_access_key": 'SECRET_KEY',
            "aws_session_token": 'SESSION_TOKEN',
        }
        c = boto3.client('ec2', **kw)
        r = boto3.resource('ec2', **kw)
        try:
            r.create_key_pair(KeyName=name)
        except NotImplementedError as e:
            raise unittest.SkipTest(
                "KeyPairs.create_key_pair not implemented in this version of moto, please update."
            ) from e
        r.create_security_group(GroupName=name, Description='the security group')
        instance = r.create_instances(ImageId=anyImageId(c), MinCount=1, MaxCount=1)[0]
        c.create_image(InstanceId=instance.id, Name="foo", Description="bar")
        c.terminate_instances(InstanceIds=[instance.id])
        return c, r

    def _patch_moto_describe_spot_price_history(self, bs, instance_type, price):
        def fake_describe_price(*args, **kwargs):
            return {'SpotPriceHistory': [{'InstanceType': instance_type, 'SpotPrice': price}]}

        self.patch(bs.ec2.meta.client, "describe_spot_price_history", fake_describe_price)

    def _patch_moto_describe_spot_instance_requests(self, c, r, bs):
        this_call = [0]

        orig_describe_instance = bs.ec2.meta.client.describe_spot_instance_requests

        def fake_describe_spot_instance_requests(*args, **kwargs):
            curr_call = this_call[0]
            this_call[0] += 1
            if curr_call == 0:
                raise ClientError(
                    {'Error': {'Code': 'InvalidSpotInstanceRequestID.NotFound'}},
                    'DescribeSpotInstanceRequests',
                )
            if curr_call == 1:
                return orig_describe_instance(*args, **kwargs)

            response = orig_describe_instance(*args, **kwargs)

            instances = r.instances.filter(
                Filters=[{'Name': 'instance-state-name', 'Values': ['running']}]
            )

            response['SpotInstanceRequests'][0]['Status']['Code'] = 'fulfilled'
            response['SpotInstanceRequests'][0]['InstanceId'] = next(iter(instances)).id
            return response

        self.patch(
            bs.ec2.meta.client,
            'describe_spot_instance_requests',
            fake_describe_spot_instance_requests,
        )

    @mock_aws
    def test_constructor_minimal(self):
        _, r = self.botoSetup('latent_buildbot_slave')
        amis = list(r.images.all())
        bs = ec2.EC2LatentWorker(
            'bot1',
            'sekrit',
            'm1.large',
            identifier='publickey',
            secret_identifier='privatekey',
            keypair_name='keypair_name',
            security_name='security_name',
            ami=amis[0].id,
        )
        self.assertEqual(bs.workername, 'bot1')
        self.assertEqual(bs.password, 'sekrit')
        self.assertEqual(bs.instance_type, 'm1.large')
        self.assertEqual(bs.ami, amis[0].id)

    @mock_aws
    def test_constructor_tags(self):
        _, r = self.botoSetup('latent_buildbot_slave')
        amis = list(r.images.all())
        tags = {'foo': 'bar'}
        bs = ec2.EC2LatentWorker(
            'bot1',
            'sekrit',
            'm1.large',
            identifier='publickey',
            secret_identifier='privatekey',
            keypair_name='keypair_name',
            security_name='security_name',
            tags=tags,
            ami=amis[0].id,
        )
        self.assertEqual(bs.tags, tags)

    @mock_aws
    def test_constructor_region(self):
        _, r = self.botoSetup()
        amis = list(r.images.all())
        bs = ec2.EC2LatentWorker(
            'bot1',
            'sekrit',
            'm1.large',
            identifier='publickey',
            secret_identifier='privatekey',
            keypair_name="latent_buildbot_worker",
            security_name='latent_buildbot_worker',
            ami=amis[0].id,
            region='us-west-1',
        )
        self.assertEqual(bs.session.region_name, 'us-west-1')

    @mock_aws
    def test_fail_mixing_classic_and_vpc_ec2_settings(self):
        _, r = self.botoSetup()
        amis = list(r.images.all())

        def create_worker():
            ec2.EC2LatentWorker(
                'bot1',
                'sekrit',
                'm1.large',
                keypair_name="test_key",
                identifier='publickey',
                secret_identifier='privatekey',
                ami=amis[0].id,
                security_name="classic",
                subnet_id="sn-1234",
            )

        with self.assertRaises(ValueError):
            create_worker()

    @mock_aws
    def test_start_vpc_instance(self):
        _, r = self.botoSetup()

        vpc = r.create_vpc(CidrBlock="192.168.0.0/24")
        subnet = r.create_subnet(VpcId=vpc.id, CidrBlock="192.168.0.0/24")
        amis = list(r.images.all())

        sg = r.create_security_group(GroupName="test_sg", Description="test_sg", VpcId=vpc.id)
        bs = ec2.EC2LatentWorker(
            'bot1',
            'sekrit',
            'm1.large',
            identifier='publickey',
            secret_identifier='privatekey',
            keypair_name="latent_buildbot_worker",
            security_group_ids=[sg.id],
            subnet_id=subnet.id,
            ami=amis[0].id,
        )
        bs._poll_resolution = 0
        instance_id, _, _ = bs._start_instance()
        instances = r.instances.filter(
            Filters=[{'Name': 'instance-state-name', 'Values': ['running']}]
        )
        instances = list(instances)

        self.assertEqual(len(instances), 1)
        self.assertEqual(instances[0].id, instance_id)
        self.assertEqual(instances[0].subnet_id, subnet.id)
        self.assertEqual(len(instances[0].security_groups), 1)
        self.assertEqual(instances[0].security_groups[0]['GroupId'], sg.id)
        self.assertEqual(instances[0].key_name, 'latent_buildbot_worker')

    @mock_aws
    def test_start_instance(self):
        _, r = self.botoSetup()
        amis = list(r.images.all())
        bs = ec2.EC2LatentWorker(
            'bot1',
            'sekrit',
            'm1.large',
            identifier='publickey',
            secret_identifier='privatekey',
            keypair_name='keypair_name',
            security_name='security_name',
            ami=amis[0].id,
        )
        bs._poll_resolution = 1
        instance_id, image_id, start_time = bs._start_instance()
        self.assertTrue(instance_id.startswith('i-'))
        self.assertTrue(image_id.startswith('ami-'))
        self.assertTrue(start_time > "00:00:00")
        instances = r.instances.filter(
            Filters=[{'Name': 'instance-state-name', 'Values': ['running']}]
        )
        instances = list(instances)
        self.assertEqual(len(instances), 1)
        self.assertEqual(instances[0].id, instance_id)
        self.assertIsNone(instances[0].tags)
        self.assertEqual(instances[0].id, bs.properties.getProperty('instance'))

    @mock_aws
    def test_start_instance_volumes(self):
        _, r = self.botoSetup()
        block_device_map_arg = [
            {
                'DeviceName': "/dev/xvdb",
                'Ebs': {
                    "VolumeType": "io1",
                    "Iops": 10,
                    "VolumeSize": 20,
                },
            },
            {
                'DeviceName': "/dev/xvdc",
                'Ebs': {
                    "VolumeType": "gp2",
                    "VolumeSize": 30,
                    "DeleteOnTermination": False,
                },
            },
        ]
        block_device_map_res = [
            {
                'DeviceName': "/dev/xvdb",
                'Ebs': {
                    "VolumeType": "io1",
                    "Iops": 10,
                    "VolumeSize": 20,
                    "DeleteOnTermination": True,
                },
            },
            {
                'DeviceName': "/dev/xvdc",
                'Ebs': {
                    "VolumeType": "gp2",
                    "VolumeSize": 30,
                    "DeleteOnTermination": False,
                },
            },
        ]

        amis = list(r.images.all())
        bs = ec2.EC2LatentWorker(
            'bot1',
            'sekrit',
            'm1.large',
            identifier='publickey',
            secret_identifier='privatekey',
            keypair_name="latent_buildbot_worker",
            security_name='latent_buildbot_worker',
            ami=amis[0].id,
            block_device_map=block_device_map_arg,
        )
        # moto does not currently map volumes properly.  below ensures
        # that my conversion code properly composes it, including
        # delete_on_termination default.
        self.assertEqual(block_device_map_res, bs.block_device_map)

    @mock_aws
    def test_start_instance_attach_volume(self):
        _, r = self.botoSetup()
        vol = r.create_volume(Size=10, AvailabilityZone='us-east-1a')
        amis = list(r.images.all())
        ami = amis[0]
        bs = ec2.EC2LatentWorker(
            'bot1',
            'sekrit',
            'm1.large',
            identifier='publickey',
            secret_identifier='privatekey',
            keypair_name="latent_buildbot_worker",
            security_name='latent_buildbot_worker',
            ami=ami.id,
            volumes=[(vol.id, "/dev/sdz")],
        )
        bs._poll_resolution = 0
        bs._start_instance()
        instances = r.instances.filter(
            Filters=[{'Name': 'instance-state-name', 'Values': ['running']}]
        )
        instances = list(instances)
        instance = instances[0]
        sdz = next(bm for bm in instance.block_device_mappings if bm['DeviceName'] == '/dev/sdz')
        self.assertEqual(vol.id, sdz['Ebs']['VolumeId'])

    @mock_aws
    def test_start_instance_tags(self):
        _, r = self.botoSetup('latent_buildbot_slave')
        amis = list(r.images.all())
        tags = {'foo': 'bar'}
        bs = ec2.EC2LatentWorker(
            'bot1',
            'sekrit',
            'm1.large',
            identifier='publickey',
            secret_identifier='privatekey',
            keypair_name="latent_buildbot_worker",
            security_name='latent_buildbot_worker',
            tags=tags,
            ami=amis[0].id,
        )
        bs._poll_resolution = 0
        id, _, _ = bs._start_instance()
        instances = r.instances.filter(
            Filters=[{'Name': 'instance-state-name', 'Values': ['running']}]
        )
        instances = list(instances)
        self.assertEqual(len(instances), 1)
        self.assertEqual(instances[0].id, id)
        self.assertEqual(instances[0].tags, [{'Value': 'bar', 'Key': 'foo'}])

    @mock_aws
    def test_start_instance_ip(self):
        c, r = self.botoSetup('latent_buildbot_slave')
        amis = list(r.images.all())
        eip = c.allocate_address(Domain='vpc')
        elastic_ip = eip['PublicIp']
        bs = ec2.EC2LatentWorker(
            'bot1',
            'sekrit',
            'm1.large',
            identifier='publickey',
            secret_identifier='privatekey',
            keypair_name="latent_buildbot_worker",
            security_name='latent_buildbot_worker',
            elastic_ip=elastic_ip,
            ami=amis[0].id,
        )
        bs._poll_resolution = 0
        bs._start_instance()
        instances = r.instances.filter(
            Filters=[{'Name': 'instance-state-name', 'Values': ['running']}]
        )
        instances = list(instances)
        addresses = c.describe_addresses()['Addresses']
        self.assertEqual(instances[0].id, addresses[0]['InstanceId'])

    @mock_aws
    def test_start_vpc_spot_instance(self):
        c, r = self.botoSetup()

        vpc = r.create_vpc(CidrBlock="192.168.0.0/24")
        subnet = r.create_subnet(VpcId=vpc.id, CidrBlock="192.168.0.0/24")
        amis = list(r.images.all())

        sg = r.create_security_group(GroupName="test_sg", Description="test_sg", VpcId=vpc.id)

        bs = ec2.EC2LatentWorker(
            'bot1',
            'sekrit',
            'm1.large',
            identifier='publickey',
            secret_identifier='privatekey',
            keypair_name="latent_buildbot_worker",
            ami=amis[0].id,
            spot_instance=True,
            max_spot_price=1.5,
            security_group_ids=[sg.id],
            subnet_id=subnet.id,
        )
        bs._poll_resolution = 0

        self._patch_moto_describe_spot_price_history(bs, 'm1.large', price=1.0)
        self._patch_moto_describe_spot_instance_requests(c, r, bs)

        instance_id, _, _ = bs._request_spot_instance()
        instances = r.instances.filter(
            Filters=[{'Name': 'instance-state-name', 'Values': ['running']}]
        )
        instances = list(instances)

        self.assertTrue(bs.spot_instance)
        self.assertEqual(len(instances), 1)
        self.assertEqual(instances[0].id, instance_id)
        self.assertEqual(instances[0].subnet_id, subnet.id)
        self.assertEqual(len(instances[0].security_groups), 1)

        # TODO: As of moto 2.0.2 GroupId is not handled in spot requests
        # self.assertEqual(instances[0].security_groups[0]['GroupId'], sg.id)

    @mock_aws
    def test_start_spot_instance(self):
        c, r = self.botoSetup('latent_buildbot_slave')
        amis = list(r.images.all())
        product_description = 'Linux/Unix'
        bs = ec2.EC2LatentWorker(
            'bot1',
            'sekrit',
            'm1.large',
            identifier='publickey',
            secret_identifier='privatekey',
            keypair_name='keypair_name',
            security_name='security_name',
            ami=amis[0].id,
            spot_instance=True,
            max_spot_price=1.5,
            product_description=product_description,
        )
        bs._poll_resolution = 0

        self._patch_moto_describe_spot_price_history(bs, 'm1.large', price=1.0)
        self._patch_moto_describe_spot_instance_requests(c, r, bs)

        instance_id, _, _ = bs._request_spot_instance()
        instances = r.instances.filter(
            Filters=[{'Name': 'instance-state-name', 'Values': ['running']}]
        )
        instances = list(instances)
        self.assertTrue(bs.spot_instance)
        self.assertEqual(bs.product_description, product_description)
        self.assertEqual(len(instances), 1)
        self.assertEqual(instances[0].id, instance_id)
        self.assertIsNone(instances[0].tags)

    @mock_aws
    def test_get_image_ami(self):
        _, r = self.botoSetup('latent_buildbot_slave')
        amis = list(r.images.all())
        ami = amis[0]
        bs = ec2.EC2LatentWorker(
            'bot1',
            'sekrit',
            'm1.large',
            identifier='publickey',
            secret_identifier='privatekey',
            keypair_name="latent_buildbot_worker",
            security_name='latent_buildbot_worker',
            ami=ami.id,
        )
        image = bs.get_image()

        self.assertEqual(image.id, ami.id)

    @mock_aws
    def test_get_image_owners(self):
        _, r = self.botoSetup('latent_buildbot_slave')
        amis = list(r.images.all())
        ami = amis[0]
        bs = ec2.EC2LatentWorker(
            'bot1',
            'sekrit',
            'm1.large',
            identifier='publickey',
            secret_identifier='privatekey',
            keypair_name="latent_buildbot_worker",
            security_name='latent_buildbot_worker',
            valid_ami_owners=[int(ami.owner_id)],
        )
        image = bs.get_image()

        self.assertEqual(image.owner_id, ami.owner_id)

    @mock_aws
    def test_get_image_location(self):
        self.botoSetup('latent_buildbot_slave')
        bs = ec2.EC2LatentWorker(
            'bot1',
            'sekrit',
            'm1.large',
            identifier='publickey',
            secret_identifier='privatekey',
            keypair_name="latent_buildbot_worker",
            security_name='latent_buildbot_worker',
            valid_ami_location_regex='amazon/.*',
        )
        image = bs.get_image()

        self.assertTrue(image.image_location.startswith("amazon/"))

    @mock_aws
    def test_get_image_location_not_found(self):
        def create_worker():
            ec2.EC2LatentWorker(
                'bot1',
                'sekrit',
                'm1.large',
                identifier='publickey',
                secret_identifier='privatekey',
                keypair_name="latent_buildbot_worker",
                security_name='latent_buildbot_worker',
                valid_ami_location_regex='foobar.*',
            )

        with self.assertRaises(ValueError):
            create_worker()

    @mock_aws
    def test_fail_multiplier_and_max_are_none(self):
        """
        price_multiplier and max_spot_price may not be None at the same time.
        """
        _, r = self.botoSetup()
        amis = list(r.images.all())

        def create_worker():
            ec2.EC2LatentWorker(
                'bot1',
                'sekrit',
                'm1.large',
                identifier='publickey',
                secret_identifier='privatekey',
                keypair_name="latent_buildbot_worker",
                security_name='latent_buildbot_worker',
                ami=amis[0].id,
                region='us-west-1',
                spot_instance=True,
                price_multiplier=None,
                max_spot_price=None,
            )

        with self.assertRaises(ValueError):
            create_worker()


class TestEC2LatentWorkerDefaultKeyairSecurityGroup(unittest.TestCase):
    ec2_connection = None

    def setUp(self):
        super().setUp()
        if boto3 is None:
            raise unittest.SkipTest("moto not found")

    def botoSetup(self):
        c = boto3.client('ec2', region_name='us-east-1')
        r = boto3.resource('ec2', region_name='us-east-1')
        try:
            r.create_key_pair(KeyName='latent_buildbot_slave')
            r.create_key_pair(KeyName='test_keypair')
        except NotImplementedError as e:
            raise unittest.SkipTest(
                "KeyPairs.create_key_pair not implemented in this version of moto, please update."
            ) from e
        r.create_security_group(GroupName='latent_buildbot_slave', Description='the security group')
        r.create_security_group(GroupName='test_security_group', Description='other security group')
        instance = r.create_instances(ImageId=anyImageId(c), MinCount=1, MaxCount=1)[0]
        c.create_image(InstanceId=instance.id, Name="foo", Description="bar")
        c.terminate_instances(InstanceIds=[instance.id])
        return c, r

    @mock_aws
    def test_no_default_security_warning_when_security_group_ids(self):
        _, r = self.botoSetup()
        amis = list(r.images.all())

        bs = ec2.EC2LatentWorker(
            'bot1',
            'sekrit',
            'm1.large',
            identifier='publickey',
            secret_identifier='privatekey',
            ami=amis[0].id,
            keypair_name='test_keypair',
            subnet_id=["sn-1"],
        )
        self.assertEqual(bs.security_name, None)

    @mock_aws
    def test_use_non_default_keypair_security(self):
        _, r = self.botoSetup()
        amis = list(r.images.all())
        with assertNotProducesWarnings(DeprecatedApiWarning):
            bs = ec2.EC2LatentWorker(
                'bot1',
                'sekrit',
                'm1.large',
                identifier='publickey',
                secret_identifier='privatekey',
                ami=amis[0].id,
                security_name='test_security_group',
                keypair_name='test_keypair',
            )
        self.assertEqual(bs.keypair_name, 'test_keypair')
        self.assertEqual(bs.security_name, 'test_security_group')



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import os
import platform
import signal

from twisted.internet import defer
from twisted.internet import reactor

from buildbot.scripts.logwatcher import BuildmasterTimeoutError
from buildbot.scripts.logwatcher import LogWatcher
from buildbot.scripts.logwatcher import ReconfigError
from buildbot.util import in_reactor
from buildbot.util import rewrap


class Reconfigurator:
    @defer.inlineCallbacks
    def run(self, basedir, quiet, timeout=None):
        # Returns "Microsoft" for Vista and "Windows" for other versions
        if platform.system() in ("Windows", "Microsoft"):
            print("Reconfig (through SIGHUP) is not supported on Windows.")
            return None

        with open(os.path.join(basedir, "twistd.pid"), encoding='utf-8') as f:
            self.pid = int(f.read().strip())
        if quiet:
            os.kill(self.pid, signal.SIGHUP)
            return None

        # keep reading twistd.log. Display all messages between "loading
        # configuration from ..." and "configuration update complete" or
        # "I will keep using the previous config file instead.", or until
        # `timeout` seconds have elapsed.

        self.sent_signal = False
        reactor.callLater(0.2, self.sighup)

        lw = LogWatcher(os.path.join(basedir, "twistd.log"), timeout=timeout)

        try:
            yield lw.start()
            print("Reconfiguration appears to have completed successfully")
            return 0
        except BuildmasterTimeoutError:
            print("Never saw reconfiguration finish.")
        except ReconfigError:
            print(
                rewrap("""\
                Reconfiguration failed. Please inspect the master.cfg file for
                errors, correct them, then try 'buildbot reconfig' again.
                """)
            )
        except OSError:
            # we were probably unable to open the file in the first place
            self.sighup()
        except Exception as e:
            print(f"Error while following twistd.log: {e}")

        return 1

    def sighup(self):
        if self.sent_signal:
            return
        print(f"sending SIGHUP to process {self.pid}")
        self.sent_signal = True
        os.kill(self.pid, signal.SIGHUP)


@in_reactor
def reconfig(config):
    basedir = config['basedir']
    quiet = config['quiet']

    timeout = config.get('progress_timeout', None)
    if timeout is not None:
        try:
            timeout = float(timeout)
        except ValueError:
            print('Progress timeout must be a number')
            return 1

    r = Reconfigurator()
    return r.run(basedir, quiet, timeout=timeout)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import html
import time
from typing import TYPE_CHECKING
from typing import Any

from twisted.internet import defer
from twisted.python import log

from buildbot import util
from buildbot.process.properties import Properties
from buildbot.util import datetime2epoch

if TYPE_CHECKING:
    from buildbot.db.changes import ChangeModel


class Change:
    """I represent a single change to the source tree. This may involve several
    files, but they are all changed by the same person, and there is a change
    comment for the group as a whole."""

    number: int | None = None
    branch: str | None = None
    category: str | None = None
    revision: str | None = None  # used to create a source-stamp
    links: list[str] = []  # links are gone, but upgrade code expects this attribute

    @classmethod
    def fromChdict(cls, master: Any, chdict: ChangeModel) -> Change:
        """
        Class method to create a L{Change} from a L{ChangeModel} as returned
        by L{ChangesConnectorComponent.getChange}.

        @param master: build master instance
        @param chdict: change model

        @returns: L{Change} via Deferred
        """
        cache = master.caches.get_cache("Changes", cls._make_ch)
        return cache.get(chdict.changeid, chdict=chdict, master=master)

    @classmethod
    def _make_ch(cls, changeid: int, master: Any, chdict: ChangeModel) -> defer.Deferred[Change]:
        change = cls(None, None, None, _fromChdict=True)
        change.who = chdict.author
        change.committer = chdict.committer
        change.comments = chdict.comments
        change.revision = chdict.revision
        change.branch = chdict.branch
        change.category = chdict.category
        change.revlink = chdict.revlink or ""
        change.repository = chdict.repository
        change.codebase = chdict.codebase
        change.project = chdict.project
        change.number = chdict.changeid

        when = chdict.when_timestamp
        if when:
            when = datetime2epoch(when)
        change.when = when

        change.files = sorted(chdict.files)

        change.properties = Properties()
        for n, (v, s) in chdict.properties.items():
            change.properties.setProperty(n, v, s)

        return defer.succeed(change)

    def __init__(
        self,
        who: str | None,
        files: list[str] | None,
        comments: str | None,
        committer: str | None = None,
        revision: str | None = None,
        when: int | None = None,
        branch: str | None = None,
        category: str | None = None,
        revlink: str = '',
        properties: dict[str, Any] | None = None,
        repository: str = '',
        codebase: str = '',
        project: str = '',
        _fromChdict: bool = False,
    ) -> None:
        if properties is None:
            properties = {}
        # skip all this madness if we're being built from the database
        if _fromChdict:
            return

        self.who = who
        self.committer = committer
        self.comments = comments

        def none_or_unicode(x: Any) -> str | None:
            if x is None:
                return x
            return str(x)

        self.revision = none_or_unicode(revision)
        now = util.now()
        if when is None:
            self.when = now
        elif when > now:
            # this happens when the committing system has an incorrect clock, for example.
            # handle it gracefully
            log.msg("received a Change with when > now; assuming the change happened now")
            self.when = now
        else:
            self.when = when
        self.branch = none_or_unicode(branch)
        self.category = none_or_unicode(category)
        self.revlink = revlink
        self.properties = Properties()
        self.properties.update(properties, "Change")
        self.repository = repository
        self.codebase = codebase
        self.project = project

        # keep a sorted list of the files, for easier display
        self.files = sorted(files or [])

    def __setstate__(self, dict: dict[str, Any]) -> None:
        self.__dict__ = dict
        # Older Changes won't have a 'properties' attribute in them
        if not hasattr(self, 'properties'):
            self.properties = Properties()
        if not hasattr(self, 'revlink'):
            self.revlink = ""

    def __str__(self) -> str:
        return (
            "Change(revision=%r, who=%r, committer=%r, branch=%r, comments=%r, "
            + "when=%r, category=%r, project=%r, repository=%r, "
            + "codebase=%r)"
        ) % (
            self.revision,
            self.who,
            self.committer,
            self.branch,
            self.comments,
            self.when,
            self.category,
            self.project,
            self.repository,
            self.codebase,
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Change):
            raise NotImplementedError
        return self.number == other.number

    def __ne__(self, other: object) -> bool:
        if not isinstance(other, Change):
            raise NotImplementedError
        return self.number != other.number

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, Change):
            raise NotImplementedError
        if self.number is None:
            return False
        if other.number is None:
            return False
        return self.number < other.number

    def __le__(self, other: object) -> bool:
        if not isinstance(other, Change):
            raise NotImplementedError
        if self.number is None:
            return other.number is None
        if other.number is None:
            return False
        return self.number <= other.number

    def __gt__(self, other: object) -> bool:
        if not isinstance(other, Change):
            raise NotImplementedError
        if self.number is None:
            return False
        if other.number is None:
            return False
        return self.number > other.number

    def __ge__(self, other: object) -> bool:
        if not isinstance(other, Change):
            raise NotImplementedError
        if self.number is None:
            return other.number is None
        if other.number is None:
            return False
        return self.number >= other.number

    def asText(self) -> str:
        data = ""
        data += "Files:\n"
        for f in self.files:
            data += f" {f}\n"
        if self.repository:
            data += f"On: {self.repository}\n"
        if self.project:
            data += f"For: {self.project}\n"
        data += f"At: {self.getTime()}\n"
        data += f"Changed By: {self.who}\n"
        data += f"Committed By: {self.committer}\n"
        data += f"Comments: {self.comments}"
        data += "Properties: \n"
        for prop in self.properties.asList():
            data += f"  {prop[0]}: {prop[1]}"
        data += '\n\n'
        return data

    def asDict(self) -> dict[str, Any]:
        """returns a dictionary with suitable info for html/mail rendering"""
        files = [{"name": f} for f in self.files]
        files.sort(key=lambda a: a['name'])

        result = {
            # Constant
            'number': self.number,
            'branch': self.branch,
            'category': self.category,
            'who': self.getShortAuthor(),
            'committer': self.committer,
            'comments': self.comments,
            'revision': self.revision,
            'rev': self.revision,
            'when': self.when,
            'at': self.getTime(),
            'files': files,
            'revlink': getattr(self, 'revlink', None),
            'properties': self.properties.asList(),
            'repository': getattr(self, 'repository', None),
            'codebase': getattr(self, 'codebase', ''),
            'project': getattr(self, 'project', None),
        }
        return result

    def getShortAuthor(self) -> str | None:
        return self.who

    def getTime(self) -> str:
        if not self.when:
            return "?"
        return time.strftime("%a %d %b %Y %H:%M:%S", time.localtime(self.when))

    def getTimes(self) -> tuple[int | None, None]:
        return (self.when, None)

    def getText(self) -> list[str]:
        return [html.escape(self.who or "")]

    def getLogs(self) -> dict[str, Any]:
        return {}



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer


class StateMixin:
    # state management

    _objectid = None

    @defer.inlineCallbacks
    def getState(self, *args, **kwargs):
        # get the objectid, if not known
        if self._objectid is None:
            self._objectid = yield self.master.db.state.getObjectId(
                self.name, self.__class__.__name__
            )

        rv = yield self.master.db.state.getState(self._objectid, *args, **kwargs)
        return rv

    @defer.inlineCallbacks
    def setState(self, key, value):
        # get the objectid, if not known
        if self._objectid is None:
            self._objectid = yield self.master.db.state.getObjectId(
                self.name, self.__class__.__name__
            )

        yield self.master.db.state.setState(self._objectid, key, value)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import errno
import os
import signal
import time
from typing import Any

from twisted.trial import unittest

from buildbot_worker.scripts import stop
from buildbot_worker.test.util import compat
from buildbot_worker.test.util import misc

try:
    from unittest import mock
except ImportError:
    from unittest import mock


class TestStopWorker(misc.FileIOMixin, misc.StdoutAssertionsMixin, unittest.TestCase):
    """
    Test buildbot_worker.scripts.stop.stopWorker()
    """

    PID = 9876

    def setUp(self) -> None:
        self.setUpStdoutAssertions()

        # patch os.chdir() to do nothing
        self.patch(os, "chdir", mock.Mock())

    def test_no_pid_file(self) -> None:
        """
        test calling stopWorker() when no pid file is present
        """

        # patch open() to raise 'file not found' exception
        self.setUpOpenError(2)

        # check that stop() raises WorkerNotRunning exception
        with self.assertRaises(stop.WorkerNotRunning):
            stop.stopWorker("", False)

    @compat.skipUnlessPlatformIs("posix")
    def test_successful_stop(self) -> None:
        """
        test stopWorker() on a successful worker stop
        """

        def emulated_kill(pid: Any, sig: int) -> None:
            if sig == 0:
                # when probed if a signal can be send to the process
                # emulate that it is dead with 'No such process' error
                raise OSError(errno.ESRCH, "dummy")

        # patch open() to return a pid file
        self.setUpOpen(str(self.PID))

        # patch os.kill to emulate successful kill
        mocked_kill = mock.Mock(side_effect=emulated_kill)
        self.patch(os, "kill", mocked_kill)

        # don't waste time
        self.patch(time, "sleep", mock.Mock())

        # check that stopWorker() sends expected signal to right PID
        # and print correct message to stdout
        exit_code = stop.stopWorker("", False)
        self.assertEqual(exit_code, 0)
        mocked_kill.assert_has_calls([mock.call(self.PID, signal.SIGTERM), mock.call(self.PID, 0)])

        self.assertStdoutEqual(f"worker process {self.PID} is dead\n")

    @compat.skipUnlessPlatformIs("posix")
    def test_stop_timeout(self) -> None:
        """
        test stopWorker() when stop timeouts
        """

        # patch open() to return a pid file
        self.setUpOpen(str(self.PID))

        # patch os.kill to emulate successful kill
        mocked_kill = mock.Mock()
        self.patch(os, "kill", mocked_kill)

        # don't waste time
        self.patch(time, "sleep", mock.Mock())

        # check that stopWorker() sends expected signal to right PID
        # and print correct message to stdout
        exit_code = stop.stopWorker("", False)
        self.assertEqual(exit_code, 1)
        mocked_kill.assert_has_calls([mock.call(self.PID, signal.SIGTERM), mock.call(self.PID, 0)])

        self.assertStdoutEqual("never saw process go away\n")


class TestStop(misc.IsWorkerDirMixin, misc.StdoutAssertionsMixin, unittest.TestCase):
    """
    Test buildbot_worker.scripts.stop.stop()
    """

    config = {"basedir": "dummy", "quiet": False}

    def test_bad_basedir(self) -> None:
        """
        test calling stop() with invalid basedir path
        """

        # patch isWorkerDir() to fail
        self.setupUpIsWorkerDir(False)

        # call startCommand() and check that correct exit code is returned
        self.assertEqual(stop.stop(self.config), 1, "unexpected exit code")

        # check that isWorkerDir was called with correct argument
        self.isWorkerDir.assert_called_once_with(self.config["basedir"])

    def test_no_worker_running(self) -> None:
        """
        test calling stop() when no worker is running
        """
        self.setUpStdoutAssertions()

        # patch basedir check to always succeed
        self.setupUpIsWorkerDir(True)

        # patch stopWorker() to raise an exception
        mock_stopWorker = mock.Mock(side_effect=stop.WorkerNotRunning())
        self.patch(stop, "stopWorker", mock_stopWorker)

        exit_code = stop.stop(self.config)
        self.assertEqual(exit_code, 0)

        self.assertStdoutEqual("worker not running\n")

    def test_successful_stop(self) -> None:
        """
        test calling stop() when worker is running
        """

        # patch basedir check to always succeed
        self.setupUpIsWorkerDir(True)

        # patch stopWorker() to do nothing
        mock_stopWorker = mock.Mock(return_value=0)
        self.patch(stop, "stopWorker", mock_stopWorker)

        exit_code = stop.stop(self.config)
        self.assertEqual(exit_code, 0)
        mock_stopWorker.assert_called_once_with(
            self.config["basedir"], self.config["quiet"], "TERM"
        )

    def test_failed_stop(self) -> None:
        """
        test failing stop()
        """

        # patch basedir check to always succeed
        self.setupUpIsWorkerDir(True)

        # patch stopWorker() to do nothing
        mock_stopWorker = mock.Mock(return_value=17)
        self.patch(stop, "stopWorker", mock_stopWorker)

        exit_code = stop.stop(self.config)
        self.assertEqual(exit_code, 17)
        mock_stopWorker.assert_called_once_with(
            self.config["basedir"], self.config["quiet"], "TERM"
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import base64
from unittest import mock

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.interfaces import LatentWorkerFailedToSubstantiate
from buildbot.process.properties import Interpolate
from buildbot.process.properties import Properties
from buildbot.test.fake import fakemaster
from buildbot.test.fake import httpclientservice as fakehttpclientservice
from buildbot.test.fake.fakebuild import FakeBuildForRendering as FakeBuild
from buildbot.test.fake.fakeprotocol import FakeTrivialConnection as FakeBot
from buildbot.test.reactor import TestReactorMixin
from buildbot.util.kubeclientservice import KubeHardcodedConfig
from buildbot.worker import kubernetes


class FakeResult:
    code = 204


def mock_delete(*args):
    return defer.succeed(FakeResult())


KUBE_CTL_PROXY_FAKE = """
import time
import sys

print("Starting to serve on 127.0.0.1:" + sys.argv[2])
sys.stdout.flush()
time.sleep(1000)
"""


KUBE_CTL_PROXY_FAKE_ERROR = """
import time
import sys

print("Issue with the config!", file=sys.stderr)
sys.stderr.flush()
sys.exit(1)
"""


class TestKubernetesWorker(TestReactorMixin, unittest.TestCase):
    worker = None

    def setUp(self):
        self.setup_test_reactor()

    @defer.inlineCallbacks
    def setupWorker(self, *args, config=None, **kwargs):
        self.patch(kubernetes.KubeLatentWorker, "_generate_random_password", lambda _: "random_pw")

        if config is None:
            config = KubeHardcodedConfig(master_url="https://kube.example.com")

        worker = kubernetes.KubeLatentWorker(
            *args, masterFQDN="buildbot-master", kube_config=config, **kwargs
        )
        self.master = yield fakemaster.make_master(self, wantData=True)
        self._http = yield fakehttpclientservice.HTTPClientService.getService(
            self.master, self, "https://kube.example.com"
        )

        yield worker.setServiceParent(self.master)
        yield self.master.startService()
        self.assertTrue(config.running)
        self.addCleanup(self.master.stopService)
        return worker

    def get_expected_metadata(self):
        return {"name": "buildbot-worker-87de7e"}

    def get_expected_spec(self, image):
        return {
            "affinity": {},
            "containers": [
                {
                    "name": "buildbot-worker-87de7e",
                    "image": image,
                    "env": [
                        {"name": "BUILDMASTER", "value": "buildbot-master"},
                        {"name": "BUILDMASTER_PROTOCOL", "value": "pb"},
                        {"name": "WORKERNAME", "value": "worker"},
                        {"name": "WORKERPASS", "value": "random_pw"},
                        {"name": "BUILDMASTER_PORT", "value": "1234"},
                    ],
                    "resources": {},
                    "volumeMounts": [],
                }
            ],
            "nodeSelector": {},
            "restartPolicy": "Never",
            "volumes": [],
        }

    def test_instantiate(self):
        worker = kubernetes.KubeLatentWorker('worker')
        # class instantiation configures nothing
        self.assertEqual(getattr(worker, '_kube', None), None)

    @defer.inlineCallbacks
    def test_wrong_arg(self):
        with self.assertRaises(TypeError):
            yield self.setupWorker('worker', wrong_param='wrong_param')

    def test_service_arg(self):
        return self.setupWorker('worker')

    @defer.inlineCallbacks
    def test_builds_may_be_incompatible(self):
        worker = yield self.setupWorker('worker')
        # http is lazily created on worker substantiation
        self.assertEqual(worker.builds_may_be_incompatible, True)

    @defer.inlineCallbacks
    def test_start_service(self):
        worker = yield self.setupWorker('worker')
        # http is lazily created on worker substantiation
        self.assertNotEqual(worker._kube, None)

    def expect_pod_delete_nonexisting(self):
        self._http.expect(
            "delete",
            "/api/v1/namespaces/default/pods/buildbot-worker-87de7e",
            params={"graceperiod": 0},
            code=404,
            content_json={"message": "Pod not found", "reason": "NotFound"},
        )

    def expect_pod_delete_existing(self, image):
        self._http.expect(
            "delete",
            "/api/v1/namespaces/default/pods/buildbot-worker-87de7e",
            params={"graceperiod": 0},
            code=200,
            content_json={
                "kind": "Pod",
                "metadata": self.get_expected_metadata(),
                "spec": self.get_expected_spec(image),
            },
        )

    def expect_pod_status_not_found(self):
        self._http.expect(
            "get",
            "/api/v1/namespaces/default/pods/buildbot-worker-87de7e/status",
            code=404,
            content_json={"kind": "Status", "reason": "NotFound"},
        )

    def expect_pod_status_exists(self, image):
        self._http.expect(
            "get",
            "/api/v1/namespaces/default/pods/buildbot-worker-87de7e/status",
            code=200,
            content_json={
                "kind": "Pod",
                "metadata": self.get_expected_metadata(),
                "spec": self.get_expected_spec(image),
            },
        )

    def expect_pod_startup(self, image):
        self._http.expect(
            "post",
            "/api/v1/namespaces/default/pods",
            json={
                "apiVersion": "v1",
                "kind": "Pod",
                "metadata": self.get_expected_metadata(),
                "spec": self.get_expected_spec(image),
            },
            code=200,
            content_json={
                "kind": "Pod",
                "metadata": self.get_expected_metadata(),
                "spec": self.get_expected_spec(image),
            },
        )

    def expect_pod_startup_error(self, image):
        self._http.expect(
            "post",
            "/api/v1/namespaces/default/pods",
            json={
                "apiVersion": "v1",
                "kind": "Pod",
                "metadata": self.get_expected_metadata(),
                "spec": self.get_expected_spec(image),
            },
            code=400,
            content_json={"kind": "Status", "reason": "MissingName", "message": "need name"},
        )

    @defer.inlineCallbacks
    def test_start_worker(self):
        worker = yield self.setupWorker('worker')
        self.expect_pod_delete_nonexisting()
        self.expect_pod_status_not_found()
        self.expect_pod_startup("rendered:buildbot/buildbot-worker")
        self.expect_pod_delete_existing("rendered:buildbot/buildbot-worker")
        self.expect_pod_status_not_found()

        d = worker.substantiate(None, FakeBuild())
        worker.attached(FakeBot())
        yield d

    @defer.inlineCallbacks
    def test_start_worker_but_error(self):
        worker = yield self.setupWorker('worker')
        self.expect_pod_delete_nonexisting()
        self.expect_pod_status_not_found()
        self.expect_pod_delete_nonexisting()
        self.expect_pod_status_not_found()

        def create_pod(namespace, spec):
            raise kubernetes.KubeJsonError(400, {'message': "yeah, but no"})

        with mock.patch.object(worker, '_create_pod', create_pod):
            with self.assertRaises(LatentWorkerFailedToSubstantiate):
                yield worker.substantiate(None, FakeBuild())
        self.assertEqual(worker.instance, None)

    @defer.inlineCallbacks
    def test_start_worker_but_error_spec(self):
        worker = yield self.setupWorker('worker')

        self.expect_pod_delete_nonexisting()
        self.expect_pod_status_not_found()
        self.expect_pod_startup_error("rendered:buildbot/buildbot-worker")
        self.expect_pod_delete_nonexisting()
        self.expect_pod_status_not_found()

        with self.assertRaises(LatentWorkerFailedToSubstantiate):
            yield worker.substantiate(None, FakeBuild())
        self.assertEqual(worker.instance, None)

    @defer.inlineCallbacks
    def test_interpolate_renderables_for_new_build(self):
        build1 = Properties(img_prop="image1")
        build2 = Properties(img_prop="image2")
        worker = yield self.setupWorker('worker', image=Interpolate("%(prop:img_prop)s"))

        self.expect_pod_delete_nonexisting()
        self.expect_pod_status_not_found()
        self.expect_pod_startup("image1")
        self.expect_pod_delete_existing("image1")
        self.expect_pod_status_not_found()

        yield worker.start_instance(build1)
        yield worker.stop_instance()
        self.assertTrue((yield worker.isCompatibleWithBuild(build2)))

    @defer.inlineCallbacks
    def test_reject_incompatible_build_while_running(self):
        build1 = Properties(img_prop="image1")
        build2 = Properties(img_prop="image2")
        worker = yield self.setupWorker('worker', image=Interpolate("%(prop:img_prop)s"))

        self.expect_pod_delete_nonexisting()
        self.expect_pod_status_not_found()
        self.expect_pod_startup("image1")
        self.expect_pod_delete_existing("image1")
        self.expect_pod_status_not_found()

        yield worker.start_instance(build1)
        self.assertFalse((yield worker.isCompatibleWithBuild(build2)))
        yield worker.stop_instance()

    @defer.inlineCallbacks
    def test_start_worker_delete_non_json_response(self):
        worker = yield self.setupWorker('worker')
        self._http.expect(
            "delete",
            "/api/v1/namespaces/default/pods/buildbot-worker-87de7e",
            params={"graceperiod": 0},
            code=404,
            content="not json",
        )

        self.expect_pod_delete_nonexisting()
        self.expect_pod_status_not_found()

        with self.assertRaises(LatentWorkerFailedToSubstantiate) as e:
            yield worker.substantiate(None, FakeBuild())
        self.assertIn("Failed to decode: not json", e.exception.args[0])

    @defer.inlineCallbacks
    def test_start_worker_delete_timeout(self):
        worker = yield self.setupWorker('worker', missing_timeout=4)

        self.expect_pod_delete_existing("rendered:buildbot/buildbot-worker")
        self.expect_pod_status_exists("rendered:buildbot/buildbot-worker")
        self.expect_pod_status_exists("rendered:buildbot/buildbot-worker")
        self.expect_pod_status_exists("rendered:buildbot/buildbot-worker")
        self.expect_pod_status_exists("rendered:buildbot/buildbot-worker")
        self.expect_pod_status_exists("rendered:buildbot/buildbot-worker")

        d = worker.stop_instance()
        self.reactor.pump([0.5] * 20)
        with self.assertRaises(TimeoutError):
            yield d

    @defer.inlineCallbacks
    def test_start_worker_create_non_json_response(self):
        worker = yield self.setupWorker('worker')

        self.expect_pod_delete_nonexisting()
        self.expect_pod_status_not_found()

        expected_metadata = {"name": "buildbot-worker-87de7e"}
        expected_spec = {
            "affinity": {},
            "containers": [
                {
                    "name": "buildbot-worker-87de7e",
                    "image": "rendered:buildbot/buildbot-worker",
                    "env": [
                        {"name": "BUILDMASTER", "value": "buildbot-master"},
                        {"name": "BUILDMASTER_PROTOCOL", "value": "pb"},
                        {"name": "WORKERNAME", "value": "worker"},
                        {"name": "WORKERPASS", "value": "random_pw"},
                        {"name": "BUILDMASTER_PORT", "value": "1234"},
                    ],
                    "resources": {},
                    "volumeMounts": [],
                }
            ],
            "nodeSelector": {},
            "restartPolicy": "Never",
            "volumes": [],
        }

        self._http.expect(
            "post",
            "/api/v1/namespaces/default/pods",
            json={
                "apiVersion": "v1",
                "kind": "Pod",
                "metadata": expected_metadata,
                "spec": expected_spec,
            },
            code=200,
            content="not json",
        )
        self.expect_pod_delete_nonexisting()
        self.expect_pod_status_not_found()

        with self.assertRaises(LatentWorkerFailedToSubstantiate) as e:
            yield worker.substantiate(None, FakeBuild())
        self.assertIn("Failed to decode: not json", e.exception.args[0])

    @defer.inlineCallbacks
    def test_hardcoded_config_verify_is_forwarded(self):
        config = KubeHardcodedConfig(
            master_url="https://kube.example.com", namespace="default", verify="/path/to/pem"
        )
        worker = yield self.setupWorker('worker', config=config)

        self._http.expect(
            "delete",
            "/api/v1/namespaces/default/pods/buildbot-worker-87de7e",
            params={"graceperiod": 0},
            verify="/path/to/pem",
            code=200,
            content_json={},
        )
        self._http.expect(
            "get",
            "/api/v1/namespaces/default/pods/buildbot-worker-87de7e/status",
            verify="/path/to/pem",
            code=404,
            content_json={"kind": "Status", "reason": "NotFound"},
        )

        yield worker.stop_instance()

    @defer.inlineCallbacks
    def test_hardcoded_config_verify_headers_is_forwarded(self):
        config = KubeHardcodedConfig(
            master_url="https://kube.example.com",
            namespace="default",
            verify="/path/to/pem",
            headers={"Test": "10"},
        )
        worker = yield self.setupWorker('worker', config=config)

        self._http.expect(
            "delete",
            "/api/v1/namespaces/default/pods/buildbot-worker-87de7e",
            params={"graceperiod": 0},
            headers={'Test': '10'},
            verify="/path/to/pem",
            code=200,
            content_json={},
        )
        self._http.expect(
            "get",
            "/api/v1/namespaces/default/pods/buildbot-worker-87de7e/status",
            headers={'Test': '10'},
            verify="/path/to/pem",
            code=404,
            content_json={"kind": "Status", "reason": "NotFound"},
        )

        yield worker.stop_instance()

    @defer.inlineCallbacks
    def test_hardcoded_config_verify_bearer_token_is_rendered(self):
        config = KubeHardcodedConfig(
            master_url="https://kube.example.com",
            namespace="default",
            verify="/path/to/pem",
            bearerToken=Interpolate("%(kw:test)s", test=10),
        )
        worker = yield self.setupWorker('worker', config=config)

        self._http.expect(
            "delete",
            "/api/v1/namespaces/default/pods/buildbot-worker-87de7e",
            params={"graceperiod": 0},
            headers={"Authorization": "Bearer 10"},
            verify="/path/to/pem",
            code=200,
            content_json={},
        )
        self._http.expect(
            "get",
            "/api/v1/namespaces/default/pods/buildbot-worker-87de7e/status",
            headers={"Authorization": "Bearer 10"},
            verify="/path/to/pem",
            code=404,
            content_json={"kind": "Status", "reason": "NotFound"},
        )

        yield worker.stop_instance()

    @defer.inlineCallbacks
    def test_hardcoded_config_verify_basicAuth_is_expanded(self):
        config = KubeHardcodedConfig(
            master_url="https://kube.example.com",
            namespace="default",
            verify="/path/to/pem",
            basicAuth={'user': 'name', 'password': Interpolate("%(kw:test)s", test=10)},
        )
        worker = yield self.setupWorker('worker', config=config)

        self._http.expect(
            "delete",
            "/api/v1/namespaces/default/pods/buildbot-worker-87de7e",
            params={"graceperiod": 0},
            headers={"Authorization": "Basic " + str(base64.b64encode(b"name:10"))},
            verify="/path/to/pem",
            code=200,
            content_json={},
        )
        self._http.expect(
            "get",
            "/api/v1/namespaces/default/pods/buildbot-worker-87de7e/status",
            headers={"Authorization": "Basic " + str(base64.b64encode(b"name:10"))},
            verify="/path/to/pem",
            code=404,
            content_json={"kind": "Status", "reason": "NotFound"},
        )

        yield worker.stop_instance()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from unittest import mock

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.process.results import FAILURE
from buildbot.reporters.base import ReporterBase
from buildbot.reporters.generators.build import BuildStatusGenerator
from buildbot.reporters.generators.worker import WorkerMissingGenerator
from buildbot.reporters.message import MessageFormatter
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util.config import ConfigErrorsMixin
from buildbot.test.util.logging import LoggingMixin
from buildbot.test.util.reporter import ReporterTestMixin


class TestException(Exception):
    pass


class TestReporterBase(
    ConfigErrorsMixin, TestReactorMixin, LoggingMixin, unittest.TestCase, ReporterTestMixin
):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.setup_reporter_test()
        self.setUpLogging()
        self.master = yield fakemaster.make_master(self, wantData=True, wantDb=True, wantMq=True)

    @defer.inlineCallbacks
    def setupNotifier(self, generators):
        mn = ReporterBase(generators=generators)
        mn.sendMessage = mock.Mock(spec=mn.sendMessage)
        mn.sendMessage.return_value = "<message>"
        yield mn.setServiceParent(self.master)
        yield mn.startService()
        return mn

    @defer.inlineCallbacks
    def setup_build_message(self, **kwargs):
        build = yield self.insert_build_finished(FAILURE)
        buildset = yield self.get_inserted_buildset()

        formatter = mock.Mock(spec=MessageFormatter)
        formatter.format_message_for_build.return_value = {
            "body": "body",
            "type": "text",
            "subject": "subject",
            "extra_info": None,
        }
        formatter.want_properties = False
        formatter.want_steps = False
        formatter.want_logs = False
        formatter.want_logs_content = False
        generator = BuildStatusGenerator(message_formatter=formatter, **kwargs)

        mn = yield self.setupNotifier(generators=[generator])

        yield mn._got_event(('builds', 20, 'finished'), build)
        return (mn, build, buildset, formatter)

    def setup_mock_generator(self, events_filter):
        gen = mock.Mock()
        gen.wanted_event_keys = events_filter
        gen.generate_name = lambda: '<name>'
        return gen

    def test_check_config_raises_error_when_generators_not_list(self):
        with self.assertRaisesConfigError('generators argument must be a list'):
            ReporterBase(generators='abc')

    @defer.inlineCallbacks
    def test_buildMessage_nominal(self):
        mn, build, buildset, formatter = yield self.setup_build_message(mode=("failing",))

        formatter.format_message_for_build.assert_called_with(
            self.master, build, is_buildset=False, mode=('failing',), users=['me@foo']
        )

        report = {
            'body': 'body',
            'subject': 'subject',
            'type': 'text',
            "extra_info": None,
            'results': FAILURE,
            'builds': [build],
            "buildset": buildset,
            'users': ['me@foo'],
            'patches': [],
            'logs': [],
        }

        self.assertEqual(mn.sendMessage.call_count, 1)
        mn.sendMessage.assert_called_with([report])

    @defer.inlineCallbacks
    def test_worker_missing_sends_message(self):
        generator = WorkerMissingGenerator(workers=['myworker'])
        mn = yield self.setupNotifier(generators=[generator])

        worker_dict = {
            'name': 'myworker',
            'notify': ["workeradmin@example.org"],
            'workerinfo': {"admin": "myadmin"},
            'last_connection': "yesterday",
        }
        yield mn._got_event(('workers', 98, 'missing'), worker_dict)

        self.assertEqual(mn.sendMessage.call_count, 1)

    @defer.inlineCallbacks
    def test_generators_subscribes_events(self):
        gen1 = self.setup_mock_generator([('fake1', None, None)])

        yield self.setupNotifier(generators=[gen1])
        self.assertEqual(len(self.master.mq.qrefs), 1)
        self.assertEqual(self.master.mq.qrefs[0].filter, ('fake1', None, None))

    @defer.inlineCallbacks
    def test_generators_subscribes_equal_events_once(self):
        gen1 = self.setup_mock_generator([('fake1', None, None)])
        gen2 = self.setup_mock_generator([('fake1', None, None)])

        yield self.setupNotifier(generators=[gen1, gen2])
        self.assertEqual(len(self.master.mq.qrefs), 1)
        self.assertEqual(self.master.mq.qrefs[0].filter, ('fake1', None, None))

    @defer.inlineCallbacks
    def test_generators_subscribes_equal_different_events_once(self):
        gen1 = self.setup_mock_generator([('fake1', None, None)])
        gen2 = self.setup_mock_generator([('fake2', None, None)])

        yield self.setupNotifier(generators=[gen1, gen2])
        self.assertEqual(len(self.master.mq.qrefs), 2)
        self.assertEqual(self.master.mq.qrefs[0].filter, ('fake1', None, None))
        self.assertEqual(self.master.mq.qrefs[1].filter, ('fake2', None, None))

    @defer.inlineCallbacks
    def test_generators_unsubscribes_on_stop_service(self):
        gen1 = self.setup_mock_generator([('fake1', None, None)])

        notifier = yield self.setupNotifier(generators=[gen1])
        yield notifier.stopService()
        self.assertEqual(len(self.master.mq.qrefs), 0)

    @defer.inlineCallbacks
    def test_generators_resubscribes_on_reconfig(self):
        gen1 = self.setup_mock_generator([('fake1', None, None)])
        gen2 = self.setup_mock_generator([('fake2', None, None)])

        notifier = yield self.setupNotifier(generators=[gen1])
        self.assertEqual(len(self.master.mq.qrefs), 1)
        self.assertEqual(self.master.mq.qrefs[0].filter, ('fake1', None, None))

        yield notifier.reconfigService(generators=[gen2])
        self.assertEqual(len(self.master.mq.qrefs), 1)
        self.assertEqual(self.master.mq.qrefs[0].filter, ('fake2', None, None))

    @defer.inlineCallbacks
    def test_generator_throw_exception_on_generate(self):
        gen = self.setup_mock_generator([('fake1', None, None)])

        @defer.inlineCallbacks
        def generate_throw(*args, **kwargs):
            raise TestException()

        gen.generate = generate_throw

        notifier = yield self.setupNotifier(generators=[gen])

        yield notifier._got_event(('fake1', None, None), None)

        self.assertEqual(len(self.flushLoggedErrors(TestException)), 1)
        self.assertLogged('Got exception when handling reporter events')

    @defer.inlineCallbacks
    def test_reports_sent_in_order_despite_slow_generator(self):
        gen = self.setup_mock_generator([('builds', None, None)])

        notifier = yield self.setupNotifier(generators=[gen])

        # Handle an event when generate is slow
        gen.generate = slow_generate = mock.Mock(return_value=defer.Deferred())
        notifier._got_event(('builds', None, None), {'buildrequestid': 1})
        notifier.sendMessage.assert_not_called()

        # Then handle an event when generate is fast
        gen.generate = mock.Mock(return_value=defer.Deferred())
        gen.generate.return_value.callback(2)
        notifier._got_event(('builds', None, None), {'buildrequestid': 1})

        # sendMessage still not called
        notifier.sendMessage.assert_not_called()

        # Have the slow generate finish
        slow_generate.return_value.callback(1)

        # Now sendMessage should have been called two times
        self.assertEqual(notifier.sendMessage.call_args_list, [mock.call([1]), mock.call([2])])

    @defer.inlineCallbacks
    def test_reports_sent_in_order_despite_multiple_slow_generators(self):
        gen = self.setup_mock_generator([('buildrequests', None, None)])
        gen2 = self.setup_mock_generator([('builds', None, None)])

        notifier = yield self.setupNotifier(generators=[gen, gen2])

        # This makes it possible to mock generate calls in arbitrary order
        mock_generate_calls = {
            'buildrequests': {1: {'new': defer.Deferred()}},
            'builds': {1: {'new': defer.Deferred(), 'finished': defer.Deferred()}},
        }

        def mock_generate(_1, _2, key, msg):
            return mock_generate_calls[key[0]][msg['buildrequestid']][key[2]]

        gen.generate = mock.Mock(side_effect=mock_generate)
        gen2.generate = mock.Mock(side_effect=mock_generate)

        # Handle an event when generate is very slow
        notifier._got_event(('buildrequests', None, 'new'), {'buildrequestid': 1})

        # Handle an event when generate is also slow
        notifier._got_event(('builds', None, 'new'), {'buildrequestid': 1})

        # Handle an event when generate is fast
        mock_generate_calls['builds'][1]['finished'].callback(3)
        notifier._got_event(('builds', None, 'finished'), {'buildrequestid': 1})

        # Finish generate call for second event
        mock_generate_calls['builds'][1]['new'].callback(2)

        # sendMessage still not called
        notifier.sendMessage.assert_not_called()

        # Finish generate call for first event
        mock_generate_calls['buildrequests'][1]['new'].callback(1)

        # Now sendMessage should have been called three times in given order
        self.assertEqual(
            notifier.sendMessage.call_args_list, [mock.call([1]), mock.call([2]), mock.call([3])]
        )

    @defer.inlineCallbacks
    def test_reports_sent_in_order_and_asap_for_multiple_builds(self):
        gen = self.setup_mock_generator([('builds', None, None)])

        notifier = yield self.setupNotifier(generators=[gen])

        # This makes it possible to mock generate calls in arbitrary order
        mock_generate_calls = {
            'builds': {
                1: {'new': defer.Deferred(), 'finished': defer.Deferred()},
                2: {'new': defer.Deferred(), 'finished': defer.Deferred()},
            }
        }

        def mock_generate(_1, _2, key, msg):
            return mock_generate_calls[key[0]][msg['buildrequestid']][key[2]]

        gen.generate = mock.Mock(side_effect=mock_generate)

        # Handle an event (for first build) when generate is slow
        notifier._got_event(('builds', None, 'new'), {'buildrequestid': 1})
        notifier.sendMessage.assert_not_called()

        # Handle an event (for second build) when generate is fast
        mock_generate_calls['builds'][2]['new'].callback(21)
        notifier._got_event(('builds', None, 'new'), {'buildrequestid': 2})

        # Handle an event (for first build) when generate is fast
        mock_generate_calls['builds'][1]['finished'].callback(12)
        notifier._got_event(('builds', None, 'finished'), {'buildrequestid': 1})

        # Handle an event (for second build) when generate is fast
        mock_generate_calls['builds'][2]['finished'].callback(22)
        notifier._got_event(('builds', None, 'finished'), {'buildrequestid': 2})

        # Finish generate call for first event
        mock_generate_calls['builds'][1]['new'].callback(11)

        # Now sendMessage should have been called four times in given order
        self.assertEqual(
            notifier.sendMessage.call_args_list,
            [mock.call([21]), mock.call([22]), mock.call([11]), mock.call([12])],
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

from twisted.internet import defer

from buildbot.machine.latent import AbstractLatentMachine
from buildbot.machine.latent import States as MachineStates
from buildbot.util import service


class FakeMachineManager(service.AsyncMultiService):
    name: str | None = 'MachineManager'  # type: ignore

    @property
    def machines(self):
        return self.namedServices

    def getMachineByName(self, name):
        if name in self.machines:
            return self.machines[name]
        return None


class LatentMachineController:
    """A controller for ``ControllableLatentMachine``"""

    def __init__(self, name, **kwargs):
        self.machine = ControllableLatentMachine(name, self, **kwargs)
        self._start_deferred = None
        self._stop_deferred = None

    def start_machine(self, result):
        assert self.machine.state == MachineStates.STARTING
        d = self._start_deferred
        self._start_deferred = None
        if isinstance(result, Exception):
            d.errback(result)
        else:
            d.callback(result)

    def stop_machine(self, result=True):
        assert self.machine.state == MachineStates.STOPPING
        d = self._stop_deferred
        self._stop_deferred = None
        if isinstance(result, Exception):
            d.errback(result)
        else:
            d.callback(result)


class ControllableLatentMachine(AbstractLatentMachine):
    """A latent machine that can be controlled by tests"""

    def __init__(self, name, controller, **kwargs):
        self._controller = controller
        super().__init__(name, **kwargs)

    def start_machine(self):
        d = defer.Deferred()
        self._controller._start_deferred = d
        return d

    def stop_machine(self):
        d = defer.Deferred()
        self._controller._stop_deferred = d
        return d



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from twisted.internet import defer

from buildbot.util import service


class DebugServices(service.ReconfigurableServiceMixin, service.AsyncMultiService):
    name: str | None = 'debug_services'  # type: ignore[assignment]

    def __init__(self):
        super().__init__()

        self.debug_port = None
        self.debug_password = None
        self.debug_registration = None
        self.manhole = None

    @defer.inlineCallbacks
    def reconfigServiceWithBuildbotConfig(self, new_config):
        if new_config.manhole != self.manhole:
            if self.manhole:
                yield self.manhole.disownServiceParent()
                self.manhole = None

            if new_config.manhole:
                self.manhole = new_config.manhole
                yield self.manhole.setServiceParent(self)

        # chain up
        yield super().reconfigServiceWithBuildbotConfig(new_config)

    @defer.inlineCallbacks
    def stopService(self):
        # manhole will get stopped as a sub-service
        yield super().stopService()

        # clean up
        if self.manhole:
            self.manhole = None



# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.trial import unittest

from buildbot.test.util.config import ConfigErrorsMixin
from buildbot.www.authz import roles


class RolesFromGroups(unittest.TestCase):
    def setUp(self):
        self.roles = roles.RolesFromGroups("buildbot-")

    def test_noGroups(self):
        ret = self.roles.getRolesFromUser({"username": 'homer'})
        self.assertEqual(ret, [])

    def test_noBuildbotGroups(self):
        ret = self.roles.getRolesFromUser({"username": "homer", "groups": ["employee"]})
        self.assertEqual(ret, [])

    def test_someBuildbotGroups(self):
        ret = self.roles.getRolesFromUser({
            "username": "homer",
            "groups": ["employee", "buildbot-maintainer", "buildbot-admin"],
        })
        self.assertEqual(ret, ["maintainer", "admin"])


class RolesFromEmails(unittest.TestCase):
    def setUp(self):
        self.roles = roles.RolesFromEmails(
            employee=["homer@plant.com", "burns@plant.com"], boss=["burns@plant.com"]
        )

    def test_noUser(self):
        ret = self.roles.getRolesFromUser({"username": 'lisa', "email": 'lisa@school.com'})
        self.assertEqual(ret, [])

    def test_User1(self):
        ret = self.roles.getRolesFromUser({"username": 'homer', "email": 'homer@plant.com'})
        self.assertEqual(ret, ["employee"])

    def test_User2(self):
        ret = self.roles.getRolesFromUser({"username": 'burns', "email": 'burns@plant.com'})
        self.assertEqual(sorted(ret), ["boss", "employee"])


class RolesFromOwner(unittest.TestCase):
    def setUp(self):
        self.roles = roles.RolesFromOwner("ownerofbuild")

    def test_noOwner(self):
        ret = self.roles.getRolesFromUser({"username": 'lisa', "email": 'lisa@school.com'}, None)
        self.assertEqual(ret, [])

    def test_notOwner(self):
        ret = self.roles.getRolesFromUser(
            {"username": 'lisa', "email": 'lisa@school.com'}, "homer@plant.com"
        )
        self.assertEqual(ret, [])

    def test_owner(self):
        ret = self.roles.getRolesFromUser(
            {"username": 'homer', "email": 'homer@plant.com'}, "homer@plant.com"
        )
        self.assertEqual(ret, ["ownerofbuild"])


class RolesFromUsername(unittest.TestCase, ConfigErrorsMixin):
    def setUp(self):
        self.roles = roles.RolesFromUsername(roles=["admins"], usernames=["Admin"])
        self.roles2 = roles.RolesFromUsername(
            roles=["developers", "integrators"], usernames=["Alice", "Bob"]
        )

    def test_anonymous(self):
        ret = self.roles.getRolesFromUser({"anonymous": True})
        self.assertEqual(ret, [])

    def test_normalUser(self):
        ret = self.roles.getRolesFromUser({"username": 'Alice'})
        self.assertEqual(ret, [])

    def test_admin(self):
        ret = self.roles.getRolesFromUser({"username": 'Admin'})
        self.assertEqual(ret, ["admins"])

    def test_multipleGroups(self):
        ret = self.roles2.getRolesFromUser({"username": 'Bob'})
        self.assertEqual(ret, ["developers", "integrators"])

    def test_badUsernames(self):
        with self.assertRaisesConfigError('Usernames cannot be None'):
            roles.RolesFromUsername(roles=[], usernames=[None])



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

# This file contains scripts run by the test_runprocess tests.  Note that since
# this code runs in a different Python interpreter, it does not necessarily
# have access to any of the Buildbot source.  Functions here should be kept
# very simple!

from __future__ import annotations

import os
import select
import signal
import subprocess
import sys
import time
from typing import Callable

import psutil

# utils


def invoke_script(function, *args):
    cmd = [sys.executable, __file__, function, *list(args)]
    if os.name == 'nt':
        DETACHED_PROCESS = 0x00000008
        subprocess.Popen(
            cmd,
            shell=False,
            stdin=None,
            stdout=None,
            stderr=None,
            close_fds=True,
            creationflags=DETACHED_PROCESS,
        )
    else:
        subprocess.Popen(cmd, shell=False, stdin=None, stdout=None, stderr=None, close_fds=True)


def write_pidfile(pidfile):
    pidfile_tmp = pidfile + "~"
    f = open(pidfile_tmp, "w")
    f.write(str(os.getpid()))
    f.close()
    os.rename(pidfile_tmp, pidfile)


def sleep_forever():
    signal.alarm(110)  # die after 110 seconds
    while True:
        time.sleep(10)


script_fns: dict[str, Callable] = {}


def script(fn):
    script_fns[fn.__name__] = fn
    return fn


# scripts


@script
def write_pidfile_and_sleep():
    pidfile = sys.argv[2]
    write_pidfile(pidfile)
    sleep_forever()


@script
def spawn_child():
    parent_pidfile, child_pidfile = sys.argv[2:]
    invoke_script('write_pidfile_and_sleep', child_pidfile)
    write_pidfile(parent_pidfile)
    sleep_forever()


@script
def wait_for_pid_death_and_write_pidfile_and_sleep():
    wait_pid = int(sys.argv[2])
    pidfile = sys.argv[3]

    while psutil.pid_exists(wait_pid):
        time.sleep(0.01)

    write_pidfile(pidfile)
    sleep_forever()


@script
def double_fork():
    if os.name == 'posix':
        # when using a PTY, the child process will get SIGHUP when the
        # parent process exits, so ignore that.
        signal.signal(signal.SIGHUP, signal.SIG_IGN)

    parent_pidfile, child_pidfile = sys.argv[2:]
    parent_pid = os.getpid()

    invoke_script('wait_for_pid_death_and_write_pidfile_and_sleep', str(parent_pid), child_pidfile)
    write_pidfile(parent_pidfile)
    sys.exit(0)


@script
def assert_stdin_closed():
    # EOF counts as readable data, so we should see stdin in the readable list,
    # although it may not appear immediately, and select may return early
    bail_at = time.time() + 10
    while True:
        r, _, __ = select.select([0], [], [], 0.01)
        if r == [0]:
            return  # success!
        if time.time() > bail_at:
            raise AssertionError()  # failure :(


# make sure this process dies if necessary

if not hasattr(signal, 'alarm'):
    signal.alarm = lambda t: 0
signal.alarm(110)  # die after 110 seconds

# dispatcher

script_fns[sys.argv[1]]()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from unittest.mock import Mock

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.process.results import SUCCESS
from buildbot.reporters import utils
from buildbot.reporters.generators.buildset import BuildSetCombinedStatusGenerator
from buildbot.reporters.generators.buildset import BuildSetStatusGenerator
from buildbot.reporters.message import MessageFormatter
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util.config import ConfigErrorsMixin
from buildbot.test.util.reporter import ReporterTestMixin


class TestBuildSetGeneratorBase(
    ConfigErrorsMixin, TestReactorMixin, ReporterTestMixin, unittest.TestCase
):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.setup_reporter_test()
        self.master = yield fakemaster.make_master(self, wantData=True, wantDb=True, wantMq=True)

    @defer.inlineCallbacks
    def insert_build_finished_get_props(self, results, **kwargs):
        build = yield self.insert_build_finished(results, **kwargs)
        yield utils.getDetailsForBuild(self.master, build, want_properties=True)
        return build

    @defer.inlineCallbacks
    def setup_generator(
        self, results=SUCCESS, message=None, db_args=None, insert_build=True, **kwargs
    ):
        if message is None:
            message = {
                "body": "body",
                "type": "text",
                "subject": "subject",
                "extra_info": None,
            }
        if db_args is None:
            db_args = {}

        if insert_build:
            build = yield self.insert_build_finished_get_props(results, **db_args)
            buildset = yield self.get_inserted_buildset()
        else:
            build = None
            buildset = yield self.insert_buildset_no_builds(results, **db_args)

        formatter = Mock(spec=MessageFormatter())
        formatter.format_message_for_build.return_value = message
        formatter.format_message_for_buildset.return_value = message
        formatter.want_logs = False
        formatter.want_logs_content = False
        formatter.want_steps = False

        g = self.GENERATOR_CLASS(message_formatter=formatter, **kwargs)

        return (g, build, buildset)


class TestBuildSetGenerator(TestBuildSetGeneratorBase):
    # Note: most of the functionality of BuildSetStatusGenerator is shared with
    # BuildStatusGenerator and is tested there.

    GENERATOR_CLASS = BuildSetStatusGenerator

    @defer.inlineCallbacks
    def buildset_message(self, g, builds, buildset):
        reporter = Mock()
        reporter.getResponsibleUsersForBuild.return_value = []

        report = yield g.buildset_message(g.formatter, self.master, reporter, builds, buildset)
        return report

    @defer.inlineCallbacks
    def generate(self, g, key, build):
        reporter = Mock()
        reporter.getResponsibleUsersForBuild.return_value = []

        report = yield g.generate(self.master, reporter, key, build)
        return report

    @defer.inlineCallbacks
    def test_buildset_message_nominal(self):
        g, build, buildset = yield self.setup_generator(mode=("change",))
        report = yield self.buildset_message(g, [build], buildset)

        g.formatter.format_message_for_build.assert_called_with(
            self.master, build, is_buildset=True, mode=('change',), users=[]
        )

        self.assertEqual(
            report,
            {
                'body': 'body',
                'subject': 'subject',
                'type': 'text',
                "extra_info": None,
                'results': SUCCESS,
                'builds': [build],
                "buildset": buildset,
                'users': [],
                'patches': [],
                'logs': [],
            },
        )

    @defer.inlineCallbacks
    def test_buildset_message_no_result(self):
        g, build, buildset = yield self.setup_generator(results=None, mode=("change",))
        buildset["results"] = None
        report = yield self.buildset_message(g, [build], buildset)

        g.formatter.format_message_for_build.assert_called_with(
            self.master, build, is_buildset=True, mode=('change',), users=[]
        )

        self.assertEqual(
            report,
            {
                'body': 'body',
                'subject': 'subject',
                'type': 'text',
                "extra_info": None,
                'results': None,
                'builds': [build],
                "buildset": buildset,
                'users': [],
                'patches': [],
                'logs': [],
            },
        )

    @defer.inlineCallbacks
    def test_buildset_message_no_result_formatter_no_subject(self):
        message = {
            "body": "body",
            "type": "text",
            "subject": None,  # deprecated unspecified subject
            "extra_info": None,
        }

        g, build, buildset = yield self.setup_generator(
            results=None, message=message, mode=("change",)
        )
        buildset["results"] = None
        report = yield self.buildset_message(g, [build], buildset)

        g.formatter.format_message_for_build.assert_called_with(
            self.master, build, is_buildset=True, mode=('change',), users=[]
        )

        self.assertEqual(
            report,
            {
                'body': 'body',
                'subject': 'Buildbot not finished in Buildbot on whole buildset',
                'type': 'text',
                "extra_info": None,
                'results': None,
                'builds': [build],
                "buildset": buildset,
                'users': [],
                'patches': [],
                'logs': [],
            },
        )

    @defer.inlineCallbacks
    def test_generate_complete(self):
        g, build, buildset = yield self.setup_generator()
        report = yield self.generate(g, ('buildsets', 98, 'complete'), buildset)

        # we retrieve build data differently when processing the buildset, so adjust it to match
        del build['buildrequest']
        del build['parentbuild']
        del build['parentbuilder']

        self.assertEqual(
            report,
            {
                'body': 'body',
                'subject': 'subject',
                'type': 'text',
                "extra_info": None,
                'results': SUCCESS,
                'builds': [build],
                "buildset": buildset,
                'users': [],
                'patches': [],
                'logs': [],
            },
        )

    @defer.inlineCallbacks
    def test_generate_complete_no_builds(self):
        g, _, buildset = yield self.setup_generator(insert_build=False)
        report = yield self.generate(g, ('buildsets', 98, 'complete'), buildset)

        g.formatter.format_message_for_build.assert_not_called()
        self.assertIsNone(report)

    @defer.inlineCallbacks
    def test_generate_complete_non_matching_builder(self):
        g, _, buildset = yield self.setup_generator(builders=['non-matched'])
        report = yield self.generate(g, ('buildsets', 98, 'complete'), buildset)

        g.formatter.format_message_for_build.assert_not_called()
        self.assertIsNone(report)

    @defer.inlineCallbacks
    def test_generate_complete_non_matching_result(self):
        g, _, buildset = yield self.setup_generator(mode=('failing',))
        report = yield self.generate(g, ('buildsets', 98, 'complete'), buildset)

        g.formatter.format_message_for_build.assert_not_called()
        self.assertIsNone(report)


class TestBuildSetCombinedGenerator(TestBuildSetGeneratorBase):
    GENERATOR_CLASS = BuildSetCombinedStatusGenerator

    @defer.inlineCallbacks
    def buildset_message(self, g, buildset, builds):
        reporter = Mock()
        report = yield g.buildset_message(g.formatter, self.master, reporter, buildset, builds)
        return report

    @defer.inlineCallbacks
    def generate(self, g, key, buildset):
        report = yield g.generate(self.master, Mock(), key, buildset)
        return report

    @defer.inlineCallbacks
    def test_buildset_message_normal(self):
        g, build, buildset = yield self.setup_generator()
        report = yield self.buildset_message(g, buildset, [build])

        g.formatter.format_message_for_buildset.assert_called_with(
            self.master, buildset, [build], is_buildset=True, mode=("passing",), users=[]
        )

        # we retrieve build data differently when processing the buildset, so adjust it to match
        del build['buildrequest']
        del build['parentbuild']
        del build['parentbuilder']

        self.assertEqual(
            report,
            {
                "body": "body",
                "subject": "subject",
                "type": "text",
                "extra_info": None,
                "results": SUCCESS,
                "builds": [build],
                "buildset": buildset,
                "users": [],
                "patches": [],
                "logs": [],
            },
        )

    @defer.inlineCallbacks
    def test_buildset_message_no_builds(self):
        g, _, buildset = yield self.setup_generator(insert_build=False)
        report = yield self.buildset_message(g, buildset, [])

        g.formatter.format_message_for_buildset.assert_called_with(
            self.master, buildset, [], is_buildset=True, mode=("passing",), users=[]
        )

        self.assertEqual(
            report,
            {
                "body": "body",
                "subject": "subject",
                "type": "text",
                "extra_info": None,
                "results": SUCCESS,
                "builds": [],
                "buildset": buildset,
                "users": [],
                "patches": [],
                "logs": [],
            },
        )

    @defer.inlineCallbacks
    def test_buildset_message_no_result(self):
        g, build, buildset = yield self.setup_generator(results=None)
        buildset["results"] = None
        report = yield self.buildset_message(g, buildset, [build])

        g.formatter.format_message_for_buildset.assert_called_with(
            self.master, buildset, [build], is_buildset=True, mode=("passing",), users=[]
        )

        # we retrieve build data differently when processing the buildset, so adjust it to match
        del build['buildrequest']
        del build['parentbuild']
        del build['parentbuilder']

        self.assertEqual(
            report,
            {
                "body": "body",
                "subject": "subject",
                "type": "text",
                "extra_info": None,
                "results": None,
                "builds": [build],
                "buildset": buildset,
                "users": [],
                "patches": [],
                "logs": [],
            },
        )

    @defer.inlineCallbacks
    def test_buildset_message_no_builds_no_result(self):
        g, _, buildset = yield self.setup_generator(results=None, insert_build=False)
        buildset["results"] = None
        report = yield self.buildset_message(g, buildset, [])

        g.formatter.format_message_for_buildset.assert_called_with(
            self.master, buildset, [], is_buildset=True, mode=("passing",), users=[]
        )

        self.assertEqual(
            report,
            {
                "body": "body",
                "subject": "subject",
                "type": "text",
                "extra_info": None,
                "results": None,
                "builds": [],
                "buildset": buildset,
                "users": [],
                "patches": [],
                "logs": [],
            },
        )

    @defer.inlineCallbacks
    def test_buildset_message_no_result_formatter_no_subject(self):
        message = {
            "body": "body",
            "type": "text",
            "subject": None,  # deprecated unspecified subject
            "extra_info": None,
        }

        g, build, buildset = yield self.setup_generator(message=message)
        report = yield self.buildset_message(g, buildset, [build])

        g.formatter.format_message_for_buildset.assert_called_with(
            self.master, buildset, [build], is_buildset=True, mode=("passing",), users=[]
        )

        # we retrieve build data differently when processing the buildset, so adjust it to match
        del build['buildrequest']
        del build['parentbuild']
        del build['parentbuilder']

        self.assertEqual(
            report,
            {
                "body": "body",
                "subject": None,
                "type": "text",
                "extra_info": None,
                "results": SUCCESS,
                "builds": [build],
                "buildset": buildset,
                "users": [],
                "patches": [],
                "logs": [],
            },
        )

    @defer.inlineCallbacks
    def test_generate_complete(self):
        g, _, buildset = yield self.setup_generator(insert_build=False)
        report = yield self.generate(g, ("buildsets", 98, "complete"), buildset)

        self.assertEqual(
            report,
            {
                "body": "body",
                "subject": "subject",
                "type": "text",
                "extra_info": None,
                "results": SUCCESS,
                "builds": [],
                "buildset": buildset,
                "users": [],
                "patches": [],
                "logs": [],
            },
        )

    @defer.inlineCallbacks
    def test_generate_complete_with_builds(self):
        g, build, buildset = yield self.setup_generator(insert_build=True)
        report = yield self.generate(g, ("buildsets", 98, "complete"), buildset)

        # we retrieve build data differently when processing the buildset, so adjust it to match
        del build['buildrequest']
        del build['parentbuild']
        del build['parentbuilder']

        self.assertEqual(
            report,
            {
                "body": "body",
                "subject": "subject",
                "type": "text",
                "extra_info": None,
                "results": SUCCESS,
                "builds": [build],
                "buildset": buildset,
                "users": [],
                "patches": [],
                "logs": [],
            },
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from unittest import mock

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.data import base
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import endpoint
from buildbot.test.util.warnings import assertProducesWarnings
from buildbot.warnings import DeprecatedApiWarning


class ResourceType(TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()

    def makeResourceTypeSubclass(self, **attributes):
        attributes.setdefault('name', 'thing')
        return type('ThingResourceType', (base.ResourceType,), attributes)

    def test_sets_master(self):
        cls = self.makeResourceTypeSubclass()
        master = mock.Mock()
        inst = cls(master)
        self.assertIdentical(inst.master, master)

    def test_getEndpoints_instances_fails(self):
        ep = base.Endpoint(None, None)
        cls = self.makeResourceTypeSubclass(endpoints=[ep])
        inst = cls(None)
        with self.assertRaises(TypeError):
            inst.getEndpoints()

    def test_getEndpoints_classes(self):
        class MyEndpoint(base.Endpoint):
            pass

        cls = self.makeResourceTypeSubclass(endpoints=[MyEndpoint])
        master = mock.Mock()
        inst = cls(master)
        eps = inst.getEndpoints()
        self.assertIsInstance(eps[0], MyEndpoint)
        self.assertIdentical(eps[0].master, master)

    @defer.inlineCallbacks
    def test_produceEvent(self):
        cls = self.makeResourceTypeSubclass(
            name='singular', eventPathPatterns=["/foo/:fooid/bar/:barid"]
        )
        master = yield fakemaster.make_master(self, wantMq=True)
        master.mq.verifyMessages = False  # since this is a pretend message
        inst = cls(master)
        inst.produceEvent(
            {"fooid": 10, "barid": '20'},  # note integer vs. string
            'tested',
        )
        master.mq.assertProductions([
            (('foo', '10', 'bar', '20', 'tested'), {"fooid": 10, "barid": '20'})
        ])

    @defer.inlineCallbacks
    def test_compilePatterns(self):
        class MyResourceType(base.ResourceType):
            eventPathPatterns = [
                "/builder/:builderid/build/:number",
                "/build/:buildid",
            ]

        master = yield fakemaster.make_master(self, wantMq=True)
        master.mq.verifyMessages = False  # since this is a pretend message
        inst = MyResourceType(master)
        self.assertEqual(inst.eventPaths, ['builder/{builderid}/build/{number}', 'build/{buildid}'])

    @defer.inlineCallbacks
    def test_compilePatterns_multiline_string_deprecation(self):
        class MyResourceType(base.ResourceType):
            eventPathPatterns = """
                /builder/:builderid/build/:number
                /build/:buildid
            """

        master = yield fakemaster.make_master(self, wantMq=True)
        master.mq.verifyMessages = False  # since this is a pretend message
        with assertProducesWarnings(
            DeprecatedApiWarning,
            message_pattern='.*ResourceType.eventPathPatterns as a multiline string is deprecated.*',
        ):
            inst = MyResourceType(master)
        self.assertEqual(inst.eventPaths, ['builder/{builderid}/build/{number}', 'build/{buildid}'])


class Endpoint(endpoint.EndpointMixin, unittest.TestCase):
    class MyResourceType(base.ResourceType):
        name = "my"

    class MyEndpoint(base.Endpoint):
        pathPatterns = [
            "/my/pattern",
        ]

    endpointClass = MyEndpoint
    resourceTypeClass = MyResourceType

    @defer.inlineCallbacks
    def setUp(self):
        yield self.setUpEndpoint()

    def test_sets_master(self):
        self.assertIdentical(self.master, self.ep.master)


class ListResult(unittest.TestCase):
    def test_constructor(self):
        lr = base.ListResult([1, 2, 3], offset=10, total=20, limit=3)
        self.assertEqual(lr.data, [1, 2, 3])
        self.assertEqual(lr.offset, 10)
        self.assertEqual(lr.total, 20)
        self.assertEqual(lr.limit, 3)

    def test_repr(self):
        lr = base.ListResult([1, 2, 3], offset=10, total=20, limit=3)
        self.assertTrue(repr(lr).startswith('ListResult'))

    def test_eq(self):
        lr1 = base.ListResult([1, 2, 3], offset=10, total=20, limit=3)
        lr2 = base.ListResult([1, 2, 3], offset=20, total=30, limit=3)
        lr3 = base.ListResult([1, 2, 3], offset=20, total=30, limit=3)
        self.assertEqual(lr2, lr3)
        self.assertNotEqual(lr1, lr2)
        self.assertNotEqual(lr1, lr3)

    def test_eq_to_list(self):
        list = [1, 2, 3]
        lr1 = base.ListResult([1, 2, 3], offset=10, total=20, limit=3)
        self.assertNotEqual(lr1, list)
        lr2 = base.ListResult([1, 2, 3], offset=None, total=None, limit=None)
        self.assertEqual(lr2, list)
        lr3 = base.ListResult([1, 2, 3], total=3)
        self.assertEqual(lr3, list)
        lr4 = base.ListResult([1, 2, 3], total=4)
        self.assertNotEqual(lr4, list)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from unittest import mock

from parameterized import parameterized
from twisted.internet import defer
from twisted.trial import unittest

from buildbot.test.reactor import TestReactorMixin
from buildbot.util import poll


class TestPollerSync(TestReactorMixin, unittest.TestCase):
    @poll.method
    def poll(self):
        self.calls += 1
        if self.fail_after_running:
            raise RuntimeError('oh noes')

    def setUp(self):
        self.setup_test_reactor()
        self.master = mock.Mock()
        self.master.reactor = self.reactor

        poll.track_poll_methods()
        self.calls = 0
        self.fail_after_running = False

        def cleanup():
            poll.reset_poll_methods()
            self.assertEqual(self.reactor.getDelayedCalls(), [])

        self.addCleanup(cleanup)

    def test_call_not_started_does_nothing(self):
        self.reactor.advance(100)
        self.assertEqual(self.calls, 0)

    def test_call_when_stopped_does_nothing(self):
        self.poll()
        self.assertEqual(self.calls, 0)

    @defer.inlineCallbacks
    def test_call_when_started_forces_run(self):
        self.poll.start(interval=100, now=False)
        self.poll()
        self.reactor.advance(0)
        self.assertEqual(self.calls, 1)
        yield self.poll.stop()

    @defer.inlineCallbacks
    def test_start_with_now_forces_run_immediately(self):
        self.poll.start(interval=10, now=True)
        self.reactor.advance(0)
        self.assertEqual(self.calls, 1)
        yield self.poll.stop()

    @defer.inlineCallbacks
    def test_start_with_now_false_does_not_run(self):
        self.poll.start(interval=10, now=False)
        self.assertEqual(self.calls, 0)
        yield self.poll.stop()

    def test_stop_on_stopped_does_nothing(self):
        self.poll.start(interval=1)
        d = self.poll.stop()
        self.assertTrue(d.called)
        d = self.poll.stop()
        self.assertTrue(d.called)

    @defer.inlineCallbacks
    def test_start_twice_error(self):
        self.poll.start(interval=1)
        with self.assertRaises(AssertionError):
            self.poll.start(interval=2)
        yield self.poll.stop()

    def test_repeats_and_stops(self):
        """Polling repeats until stopped, and stop returns a Deferred"""
        self.poll.start(interval=10, now=True)
        self.reactor.advance(0)
        while self.reactor.seconds() <= 200:
            self.assertEqual(self.calls, (self.reactor.seconds() // 10) + 1)
            self.reactor.advance(1)

        d = self.poll.stop()
        self.assertTrue(d.called)

        self.assertEqual(self.calls, 21)
        self.reactor.advance(10)
        self.assertEqual(self.calls, 21)

    @defer.inlineCallbacks
    def test_fail_reschedules_and_logs_exceptions(self):
        self.fail_after_running = True
        self.poll.start(interval=1, now=True)
        self.reactor.advance(0)
        self.assertEqual(self.calls, 1)
        self.reactor.advance(1)
        self.assertEqual(self.calls, 2)
        self.assertEqual(len(self.flushLoggedErrors(RuntimeError)), 2)
        yield self.poll.stop()

    @parameterized.expand([
        ('shorter_than_interval_now_True', 5, True),
        ('longer_than_interval_now_True', 15, True),
        ('shorter_than_interval_now_False', 5, False),
        ('longer_than_interval_now_False', 15, False),
    ])
    @defer.inlineCallbacks
    def test_run_with_random_delay(self, name, random_delay_max, now):
        interval = 10

        with mock.patch("buildbot.util.poll.randint", return_value=random_delay_max):
            self.poll.start(interval=interval, now=now, random_delay_max=random_delay_max)
            self.reactor.advance(0)

            if not now:
                i = 0
                while i < interval:
                    self.assertEqual(self.calls, 0)
                    self.reactor.advance(1)
                    i += 1

            i = 0
            while i < random_delay_max:
                self.assertEqual(self.calls, 0)
                self.reactor.advance(1)
                i += 1
            self.assertEqual(self.calls, 1)
        yield self.poll.stop()

    @parameterized.expand([
        ('now_True', True),
        ('now_False', False),
    ])
    @defer.inlineCallbacks
    def test_run_with_random_delay_zero_interval_still_delays(self, name, now):
        random_delay_max = 5
        with mock.patch("buildbot.util.poll.randint", return_value=random_delay_max):
            self.poll.start(interval=0, now=now, random_delay_max=random_delay_max)
            self.reactor.advance(0)
            self.assertEqual(self.calls, 0)

            i = 0
            while i < random_delay_max:
                self.assertEqual(self.calls, 0)
                self.reactor.advance(1)
                i += 1
            self.assertEqual(self.calls, 1)

        yield self.poll.stop()

    @defer.inlineCallbacks
    def test_run_with_random_delay_stops_immediately_during_delay_phase(self):
        random_delay_max = 5
        with mock.patch("buildbot.util.poll.randint", return_value=random_delay_max):
            self.poll.start(interval=10, now=True, random_delay_max=random_delay_max)
            self.reactor.advance(1)
            self.assertEqual(self.calls, 0)
        yield self.poll.stop()


class TestPollerAsync(TestReactorMixin, unittest.TestCase):
    @poll.method
    @defer.inlineCallbacks
    def poll(self):
        assert not self.running, "overlapping call"
        self.running = True

        d = defer.Deferred()
        self.reactor.callLater(self.duration, d.callback, None)
        yield d

        self.calls += 1
        self.running = False

        if self.fail_after_running:
            raise RuntimeError('oh noes')

    def setUp(self):
        self.setup_test_reactor()
        self.master = mock.Mock()
        self.master.reactor = self.reactor

        poll.track_poll_methods()
        self.calls = 0
        self.running = False
        self.duration = 1
        self.fail_after_running = False

        self.addCleanup(poll.reset_poll_methods)

    @defer.inlineCallbacks
    def test_call_when_started_forces_run(self):
        self.poll.start(interval=10, now=True)
        self.reactor.advance(0)
        self.assertEqual(self.calls, 0)
        self.assertTrue(self.running)
        self.reactor.advance(self.duration)
        self.assertEqual(self.calls, 1)
        self.assertFalse(self.running)
        yield self.poll.stop()

    def test_repeats_and_stops(self):
        """Polling repeats until stopped, and stop returns a Deferred.  The
        duration of the function's execution does not affect the execution
        interval: executions occur every 10 seconds."""
        self.poll.start(interval=10, now=True)
        self.reactor.advance(0)

        while self.reactor.seconds() <= 200:
            self.assertEqual(self.calls, (self.reactor.seconds() + 9) // 10)
            self.assertEqual(self.running, self.reactor.seconds() % 10 == 0)
            self.reactor.advance(1)

        d = self.poll.stop()
        self.assertTrue(d.called)

        self.assertEqual(self.calls, 21)
        self.reactor.advance(10)
        self.assertEqual(self.calls, 21)

    @parameterized.expand([
        ('now_True', True),
        ('now_False', False),
    ])
    @defer.inlineCallbacks
    def test_zero_interval_starts_immediately(self, name, now):
        self.poll.start(interval=0, now=now)
        self.reactor.advance(0)

        self.assertEqual(self.calls, 0)
        self.assertTrue(self.running)
        self.reactor.advance(1)
        self.assertEqual(self.calls, 1)
        self.assertTrue(self.running)
        self.reactor.pump([1] * 10)
        self.assertEqual(self.calls, 11)
        self.assertTrue(self.running)
        d = self.poll.stop()
        self.assertTrue(self.running)
        self.reactor.advance(1)
        self.assertFalse(self.running)
        yield d

    @defer.inlineCallbacks
    def test_fail_reschedules_and_logs_exceptions(self):
        self.fail_after_running = True
        self.poll.start(interval=10, now=True)
        self.reactor.advance(0)
        self.assertTrue(self.running)
        self.reactor.advance(1)
        self.assertEqual(self.calls, 1)
        self.reactor.advance(10)
        self.assertTrue(self.running)
        self.reactor.advance(1)
        self.assertEqual(self.calls, 2)
        self.assertEqual(len(self.flushLoggedErrors(RuntimeError)), 2)
        yield self.poll.stop()

    def test_stop_while_running_waits_for_completion(self):
        self.duration = 2
        self.poll.start(interval=10)
        self.reactor.advance(0)
        self.assertFalse(self.running)
        self.reactor.advance(10)
        self.assertTrue(self.running)
        d = self.poll.stop()
        self.assertFalse(d.called)  # not stopped yet
        self.reactor.advance(1)
        self.assertFalse(d.called)
        self.reactor.advance(1)
        self.assertTrue(d.called)

    def test_call_while_waiting_schedules_immediately(self):
        self.poll.start(interval=10)
        self.reactor.advance(0)
        self.reactor.advance(5)
        self.poll()
        self.reactor.advance(0)
        self.assertTrue(self.running)
        self.reactor.advance(1)
        self.assertEqual(self.calls, 1)
        self.assertFalse(self.running)
        self.reactor.advance(4)
        self.assertTrue(self.running)
        self.reactor.advance(1)
        self.assertEqual(self.calls, 2)

    def test_call_while_running_reschedules_immediately_after(self):
        self.duration = 5
        self.poll.start(interval=10, now=True)
        self.reactor.advance(0)
        self.assertTrue(self.running)
        self.reactor.advance(3)
        self.poll()
        self.reactor.advance(2)
        self.assertEqual(self.calls, 1)
        self.reactor.advance(5)
        self.assertEqual(self.calls, 2)

    def test_call_while_running_then_stop(self):
        """Calling the poll method while the decorated method is running, then
        calling stop will not wait for both invocations to complete."""
        self.duration = 5
        self.poll.start(interval=10, now=True)
        self.reactor.advance(0)
        self.assertTrue(self.running)
        self.reactor.advance(3)
        self.assertTrue(self.running)
        self.poll()
        d = self.poll.stop()
        self.reactor.advance(2)
        self.assertEqual(self.calls, 1)
        self.assertTrue(d.called)

        self.reactor.advance(5)
        self.assertEqual(self.calls, 1)

    def test_stop_twice_while_running(self):
        """If stop is called *twice* while the poll function is running, then
        neither Deferred fires until the run is complete."""
        self.duration = 2
        self.poll.start(interval=10)
        self.reactor.advance(0)
        self.assertFalse(self.running)
        self.reactor.advance(10)
        self.assertTrue(self.running)
        d1 = self.poll.stop()
        self.assertFalse(d1.called)  # not stopped yet
        self.reactor.advance(1)
        d2 = self.poll.stop()
        self.assertFalse(d2.called)
        self.reactor.advance(1)
        self.assertTrue(d1.called)
        self.assertTrue(d2.called)

    @defer.inlineCallbacks
    def test_stop_and_restart(self):
        """If the method is immediately restarted from a callback on a stop Deferred,
        the polling continues with the new start time."""
        self.duration = 6
        self.poll.start(interval=10)
        self.reactor.advance(0)
        self.assertFalse(self.running)
        self.reactor.advance(10)
        self.assertTrue(self.running)
        d = self.poll.stop()
        self.assertFalse(d.called)  # not stopped yet
        self.reactor.advance(6)
        self.assertFalse(self.running)
        self.assertTrue(d.called)
        yield d
        self.poll.start(interval=10)
        self.reactor.advance(10)
        self.assertEqual(self.reactor.seconds(), 26)
        self.assertTrue(self.running)

        self.reactor.advance(6)
        yield self.poll.stop()

    def test_method_longer_than_interval_invoked_at_interval_multiples(self):
        self.duration = 4
        self.poll.start(interval=3, now=True)
        self.reactor.advance(0)

        exp = [
            (0, True, 0),
            (1, True, 0),
            (2, True, 0),
            (3, True, 0),
            (4, False, 1),
            (5, False, 1),
            (6, True, 1),  # next multiple of 3
            (10, False, 2),
            (12, True, 2),
            (16, False, 3),
        ]
        for secs, running, calls in exp:
            while self.reactor.seconds() < secs:
                self.reactor.advance(1)
            self.assertEqual(self.running, running)
            self.assertEqual(self.calls, calls)

    @parameterized.expand([
        ('shorter_than_interval_now_True', 5, True),
        ('longer_than_interval_now_True', 15, True),
        ('shorter_than_interval_now_False', 5, False),
        ('longer_than_interval_now_False', 15, False),
    ])
    @defer.inlineCallbacks
    def test_run_with_random_delay(self, name, random_delay_max, now):
        interval = 10

        with mock.patch("buildbot.util.poll.randint", return_value=random_delay_max):
            self.poll.start(interval=interval, now=now, random_delay_max=random_delay_max)
            self.reactor.advance(0)

            if not now:
                i = 0
                while i < interval:
                    self.assertFalse(self.running)
                    self.assertEqual(self.calls, 0)
                    self.reactor.advance(1)
                    i += 1

            i = 0
            while i < random_delay_max:
                self.assertFalse(self.running)
                self.assertEqual(self.calls, 0)
                self.reactor.advance(1)
                i += 1

            self.assertEqual(self.calls, 0)
            self.assertTrue(self.running)
            self.reactor.advance(self.duration)
            self.assertEqual(self.calls, 1)
            self.assertFalse(self.running)
        yield self.poll.stop()

    @parameterized.expand([
        ('now_True', True),
        ('now_False', False),
    ])
    @defer.inlineCallbacks
    def test_run_with_random_delay_zero_interval_still_delays(self, name, now):
        random_delay_max = 5
        with mock.patch("buildbot.util.poll.randint", return_value=random_delay_max):
            self.poll.start(interval=0, now=now, random_delay_max=random_delay_max)
            self.reactor.advance(0)
            self.assertFalse(self.running)
            self.assertEqual(self.calls, 0)

            i = 0
            while i < random_delay_max:
                self.assertFalse(self.running)
                self.assertEqual(self.calls, 0)
                self.reactor.advance(1)
                i += 1

            self.assertTrue(self.running)
            self.reactor.advance(1)
            self.assertEqual(self.calls, 1)
            self.assertFalse(self.running)

        yield self.poll.stop()

    @defer.inlineCallbacks
    def test_run_with_random_delay_stops_immediately_during_delay_phase(self):
        random_delay_max = 5
        with mock.patch("buildbot.util.poll.randint", return_value=random_delay_max):
            self.poll.start(interval=10, now=True, random_delay_max=random_delay_max)
            self.reactor.advance(1)
            self.assertFalse(self.running)
            self.assertEqual(self.calls, 0)
        yield self.poll.stop()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.trial import unittest

from buildbot.configurators import ConfiguratorBase
from buildbot.test.util import configurators


class ConfiguratorBaseTests(configurators.ConfiguratorMixin, unittest.SynchronousTestCase):
    ConfiguratorClass = ConfiguratorBase

    def test_basic(self):
        self.setupConfigurator()
        self.assertEqual(
            self.config_dict,
            {
                'schedulers': [],
                'protocols': {},
                'builders': [],
                'workers': [],
                'projects': [],
                'secretsProviders': [],
                'www': {},
            },
        )
        self.assertEqual(self.configurator.workers, [])



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from typing import ClassVar
from typing import Sequence

from twisted.trial import unittest

from buildbot import util


class ComparableMixin(unittest.TestCase):
    class Foo(util.ComparableMixin):
        compare_attrs: ClassVar[Sequence[str]] = ("a", "b")

        def __init__(self, a, b, c):
            self.a, self.b, self.c = a, b, c

    class Bar(Foo, util.ComparableMixin):
        compare_attrs: ClassVar[Sequence[str]] = ("b", "c")

    def setUp(self):
        self.f123 = self.Foo(1, 2, 3)
        self.f124 = self.Foo(1, 2, 4)
        self.f134 = self.Foo(1, 3, 4)
        self.b123 = self.Bar(1, 2, 3)
        self.b223 = self.Bar(2, 2, 3)
        self.b213 = self.Bar(2, 1, 3)

    def test_equality_identity(self):
        self.assertEqual(self.f123, self.f123)

    def test_equality_same(self):
        another_f123 = self.Foo(1, 2, 3)
        self.assertEqual(self.f123, another_f123)

    def test_equality_unimportantDifferences(self):
        self.assertEqual(self.f123, self.f124)

    def test_inequality_unimportantDifferences_subclass(self):
        # verify that the parent class's compare_attrs does
        # affect the subclass
        self.assertNotEqual(self.b123, self.b223)

    def test_inequality_importantDifferences(self):
        self.assertNotEqual(self.f123, self.f134)

    def test_inequality_importantDifferences_subclass(self):
        self.assertNotEqual(self.b123, self.b213)

    def test_inequality_differentClasses(self):
        self.assertNotEqual(self.f123, self.b123)

    def test_instance_attribute_not_used(self):
        # setting compare_attrs as an instance method doesn't
        # affect the outcome of the comparison
        another_f123 = self.Foo(1, 2, 3)
        another_f123.compare_attrs = ("b", "a")  # type: ignore
        self.assertEqual(self.f123, another_f123)

    def test_ne_importantDifferences(self):
        self.assertNotEqual(self.f123, self.f134)

    def test_ne_differentClasses(self):
        self.assertNotEqual(self.f123, self.b123)

    def test_compare(self):
        self.assertEqual(self.f123, self.f123)
        self.assertNotEqual(self.b223, self.b213)
        self.assertGreater(self.b223, self.b213)

        # Different classes
        self.assertFalse(self.b223 > self.f123)

        self.assertGreaterEqual(self.b223, self.b213)
        self.assertGreaterEqual(self.b223, self.b223)

        # Different classes
        self.assertFalse(self.f123 >= self.b123)

        self.assertLess(self.b213, self.b223)
        self.assertLessEqual(self.b213, self.b223)
        self.assertLessEqual(self.b213, self.b213)

        # Different classes
        self.assertFalse(self.f123 <= self.b123)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any

from twisted.internet import defer

from buildbot.data import base
from buildbot.data import types
from buildbot.db.buildrequests import AlreadyClaimedError
from buildbot.db.buildrequests import NotClaimedError
from buildbot.process.results import RETRY

if TYPE_CHECKING:
    import datetime
    from typing import Sequence

    from buildbot.data.resultspec import ResultSpec
    from buildbot.db.buildrequests import BuildRequestModel


def _db2data(dbmodel: BuildRequestModel, properties: dict | None):
    return {
        'buildrequestid': dbmodel.buildrequestid,
        'buildsetid': dbmodel.buildsetid,
        'builderid': dbmodel.builderid,
        'priority': dbmodel.priority,
        'claimed': dbmodel.claimed,
        'claimed_at': dbmodel.claimed_at,
        'claimed_by_masterid': dbmodel.claimed_by_masterid,
        'complete': dbmodel.complete,
        'results': dbmodel.results,
        'submitted_at': dbmodel.submitted_at,
        'complete_at': dbmodel.complete_at,
        'waited_for': dbmodel.waited_for,
        'properties': properties,
    }


def _generate_filtered_properties(props: dict, filters: Sequence) -> dict | None:
    """
    This method returns Build's properties according to property filters.

    :param props: Properties as a dict (from db)
    :param filters: Desired properties keys as a list (from API URI)
    """
    # by default no properties are returned
    if not props and not filters:
        return None

    set_filters = set(filters)
    if '*' in set_filters:
        return props

    return {k: v for k, v in props.items() if k in set_filters}


buildrequests_field_mapping = {
    'buildrequestid': 'buildrequests.id',
    'buildsetid': 'buildrequests.buildsetid',
    'builderid': 'buildrequests.builderid',
    'priority': 'buildrequests.priority',
    'complete': 'buildrequests.complete',
    'results': 'buildrequests.results',
    'submitted_at': 'buildrequests.submitted_at',
    'complete_at': 'buildrequests.complete_at',
    'waited_for': 'buildrequests.waited_for',
    # br claim
    'claimed_at': 'buildrequest_claims.claimed_at',
    'claimed_by_masterid': 'buildrequest_claims.masterid',
}


@defer.inlineCallbacks
def _get_buildset_properties_filtered(master, buildsetid: int, filters: Sequence):
    if not filters:
        return None

    props = yield master.db.buildsets.getBuildsetProperties(buildsetid)
    return _generate_filtered_properties(props, filters)


class BuildRequestEndpoint(base.Endpoint):
    kind = base.EndpointKind.SINGLE
    pathPatterns = [
        "/buildrequests/n:buildrequestid",
    ]

    @defer.inlineCallbacks
    def get(self, resultSpec: ResultSpec, kwargs):
        buildrequest = yield self.master.db.buildrequests.getBuildRequest(kwargs['buildrequestid'])
        if not buildrequest:
            return None

        filters = resultSpec.popProperties() if hasattr(resultSpec, 'popProperties') else []
        properties = yield _get_buildset_properties_filtered(
            self.master, buildrequest.buildsetid, filters
        )
        return _db2data(buildrequest, properties)

    @defer.inlineCallbacks
    def set_request_priority(self, brid, args, kwargs):
        priority = args['priority']
        yield self.master.db.buildrequests.set_build_requests_priority(
            brids=[brid], priority=priority
        )

    @defer.inlineCallbacks
    def control(self, action, args, kwargs):
        brid = kwargs['buildrequestid']
        if action == "cancel":
            self.master.mq.produce(
                ('control', 'buildrequests', str(brid), 'cancel'),
                {'reason': args.get('reason', 'no reason')},
            )
        elif action == "set_priority":
            yield self.set_request_priority(brid, args, kwargs)
        else:
            raise ValueError(f"action: {action} is not supported")


class BuildRequestsEndpoint(base.Endpoint):
    kind = base.EndpointKind.COLLECTION
    pathPatterns = [
        "/buildrequests",
        "/builders/n:builderid/buildrequests",
    ]
    rootLinkName = 'buildrequests'

    @defer.inlineCallbacks
    def get(self, resultSpec, kwargs):
        builderid = kwargs.get("builderid", None)
        complete = resultSpec.popBooleanFilter('complete')
        claimed_by_masterid = resultSpec.popBooleanFilter('claimed_by_masterid')
        if claimed_by_masterid:
            # claimed_by_masterid takes precedence over 'claimed' filter
            # (no need to check consistency with 'claimed' filter even if
            # 'claimed'=False with 'claimed_by_masterid' set, doesn't make sense)
            claimed = claimed_by_masterid
        else:
            claimed = resultSpec.popBooleanFilter('claimed')

        bsid = resultSpec.popOneFilter('buildsetid', 'eq')
        resultSpec.fieldMapping = buildrequests_field_mapping
        buildrequests = yield self.master.db.buildrequests.getBuildRequests(
            builderid=builderid,
            complete=complete,
            claimed=claimed,
            bsid=bsid,
            resultSpec=resultSpec,
        )
        results = []
        filters = resultSpec.popProperties() if hasattr(resultSpec, 'popProperties') else []
        for br in buildrequests:
            properties = yield _get_buildset_properties_filtered(
                self.master, br.buildsetid, filters
            )
            results.append(_db2data(br, properties))
        return results


class BuildRequest(base.ResourceType):
    name = "buildrequest"
    plural = "buildrequests"
    endpoints = [BuildRequestEndpoint, BuildRequestsEndpoint]
    eventPathPatterns = [
        "/buildsets/:buildsetid/builders/:builderid/buildrequests/:buildrequestid",
        "/buildrequests/:buildrequestid",
        "/builders/:builderid/buildrequests/:buildrequestid",
    ]

    class EntityType(types.Entity):
        buildrequestid = types.Integer()
        buildsetid = types.Integer()
        builderid = types.Integer()
        priority = types.Integer()
        claimed = types.Boolean()
        claimed_at = types.NoneOk(types.DateTime())
        claimed_by_masterid = types.NoneOk(types.Integer())
        complete = types.Boolean()
        results = types.NoneOk(types.Integer())
        submitted_at = types.DateTime()
        complete_at = types.NoneOk(types.DateTime())
        waited_for = types.Boolean()
        properties = types.NoneOk(types.SourcedProperties())

    entityType = EntityType(name)

    @defer.inlineCallbacks
    def generateEvent(self, brids, event):
        events = []
        for brid in brids:
            # get the build and munge the result for the notification
            br = yield self.master.data.get(('buildrequests', str(brid)))
            events.append(br)
        for br in events:
            self.produceEvent(br, event)

    @defer.inlineCallbacks
    def callDbBuildRequests(self, brids, db_callable, event, **kw):
        if not brids:
            # empty buildrequest list. No need to call db API
            return True
        try:
            yield db_callable(brids, **kw)
        except AlreadyClaimedError:
            # the db layer returned an AlreadyClaimedError exception, usually
            # because one of the buildrequests has already been claimed by
            # another master
            return False
        yield self.generateEvent(brids, event)
        return True

    @base.updateMethod
    def claimBuildRequests(self, brids: list[int], claimed_at: datetime.datetime | None = None):
        return self.callDbBuildRequests(
            brids,
            self.master.db.buildrequests.claimBuildRequests,
            event="claimed",
            claimed_at=claimed_at,
        )

    @base.updateMethod
    @defer.inlineCallbacks
    def unclaimBuildRequests(self, brids: list[int]):
        if brids:
            yield self.master.db.buildrequests.unclaimBuildRequests(brids)
            yield self.generateEvent(brids, "unclaimed")

    @base.updateMethod
    @defer.inlineCallbacks
    def completeBuildRequests(
        self, brids: list[int], results: int, complete_at: datetime.datetime | None = None
    ):
        assert results != RETRY, "a buildrequest cannot be completed with a retry status!"
        if not brids:
            # empty buildrequest list. No need to call db API
            return True
        try:
            yield self.master.db.buildrequests.completeBuildRequests(
                brids, results, complete_at=complete_at
            )
        except NotClaimedError:
            # the db layer returned a NotClaimedError exception, usually
            # because one of the buildrequests has been claimed by another
            # master
            return False
        yield self.generateEvent(brids, "complete")

        # check for completed buildsets -- one call for each build request with
        # a unique bsid
        seen_bsids = set()
        for brid in brids:
            brdict = yield self.master.db.buildrequests.getBuildRequest(brid)

            if brdict:
                bsid = brdict.buildsetid
                if bsid in seen_bsids:
                    continue
                seen_bsids.add(bsid)
                yield self.master.data.updates.maybeBuildsetComplete(bsid)

        return True

    @base.updateMethod
    @defer.inlineCallbacks
    def rebuildBuildrequest(self, buildrequest: dict[str, Any]):
        # goal is to make a copy of the original buildset
        buildset = yield self.master.data.get(('buildsets', buildrequest['buildsetid']))
        properties = yield self.master.data.get((
            'buildsets',
            buildrequest['buildsetid'],
            'properties',
        ))
        # use original build id: after rebuild, it is saved in new buildset `rebuilt_buildid` column
        builds = yield self.master.data.get((
            'buildrequests',
            buildrequest['buildrequestid'],
            'builds',
        ))
        # if already rebuilt build of the same initial build is rebuilt again only save the build
        # id of the initial build
        if len(builds) != 0 and buildset['rebuilt_buildid'] is None:
            rebuilt_buildid = builds[0]['buildid']
        else:
            rebuilt_buildid = buildset['rebuilt_buildid']

        ssids = [ss['ssid'] for ss in buildset['sourcestamps']]
        res = yield self.master.data.updates.addBuildset(
            waited_for=False,
            scheduler='rebuild',
            sourcestamps=ssids,
            reason='rebuild',
            properties=properties,
            builderids=[buildrequest['builderid']],
            external_idstring=buildset['external_idstring'],
            rebuilt_buildid=rebuilt_buildid,
            parent_buildid=buildset['parent_buildid'],
            parent_relationship=buildset['parent_relationship'],
        )
        return res



"""
Tests for `buildbot_worker.util._hangcheck`.
"""

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import cast

from twisted.internet import defer
from twisted.internet import reactor
from twisted.internet.endpoints import TCP4ClientEndpoint
from twisted.internet.endpoints import TCP4ServerEndpoint
from twisted.internet.error import ConnectionDone
from twisted.internet.protocol import Protocol
from twisted.internet.task import Clock
from twisted.internet.testing import StringTransport
from twisted.python.failure import Failure
from twisted.spread.pb import PBClientFactory
from twisted.trial.unittest import TestCase
from twisted.web.static import Data

from buildbot_worker.test.util.site import SiteWithClose

from ..util import HangCheckFactory
from ..util._hangcheck import HangCheckProtocol

try:
    from twisted.internet.testing import AccumulatingProtocol
    from twisted.internet.testing import StringTransport
except ImportError:
    from twisted.test.proto_helpers import AccumulatingProtocol
    from twisted.test.proto_helpers import StringTransport

if TYPE_CHECKING:
    from twisted.internet.interfaces import IReactorTime

    from buildbot_worker.util.twisted import InlineCallbacksType


def assert_clock_idle(case: TestCase, clock: IReactorTime) -> None:
    """
    Assert that the given clock doesn't have any pending delayed
    calls.
    """
    case.assertEqual(
        clock.getDelayedCalls(),
        [],
    )


class HangCheckTests(TestCase):
    """
    Tests for HangCheckProtocol.
    """

    # On slower machines with high CPU oversubscription this test may take longer to run than
    # the default timeout.
    timeout = 20

    def test_disconnects(self) -> None:
        """
        When connecting to a server that doesn't send any data,
        the protocol disconnects after the timeout.
        """
        clock = Clock()
        protocol = HangCheckProtocol(Protocol(), reactor=clock)
        transport = StringTransport()

        transport.protocol = protocol  # type: ignore[attr-defined]
        protocol.makeConnection(transport)

        clock.advance(120)
        self.assertTrue(transport.disconnecting)
        assert_clock_idle(self, clock)

    def test_transport(self) -> None:
        """
        The transport passed to the underlying protocol is
        the underlying transport.
        """
        clock = Clock()
        wrapped_protocol = Protocol()
        protocol = HangCheckProtocol(wrapped_protocol, reactor=clock)
        transport = StringTransport()

        transport.protocol = protocol  # type: ignore[attr-defined]
        protocol.makeConnection(transport)

        self.assertIdentical(wrapped_protocol.transport, transport)

    def test_forwards_data(self) -> None:
        """
        Data received by the protocol gets passed to the wrapped
        protocol.
        """
        clock = Clock()
        wrapped_protocol = AccumulatingProtocol()
        protocol = HangCheckProtocol(wrapped_protocol, reactor=clock)
        transport = StringTransport()

        transport.protocol = protocol  # type: ignore[attr-defined]
        protocol.makeConnection(transport)
        protocol.dataReceived(b'some-data')

        self.assertEqual(wrapped_protocol.data, b"some-data")

    def test_data_cancels_timeout(self) -> None:
        """
        When data is received, the timeout is canceled
        """
        clock = Clock()
        protocol = HangCheckProtocol(Protocol(), reactor=clock)
        transport = StringTransport()

        transport.protocol = protocol  # type: ignore[attr-defined]
        protocol.makeConnection(transport)
        protocol.dataReceived(b'some-data')
        assert_clock_idle(self, clock)

    def test_calls_callback(self) -> None:
        """
        When connecting to a server that doesn't send any data,
        the protocol calls the hung callback.
        """
        results = []
        clock = Clock()
        protocol = HangCheckProtocol(
            Protocol(),
            hung_callback=lambda: results.append(True),
            reactor=clock,
        )
        transport = StringTransport()

        transport.protocol = protocol  # type: ignore[attr-defined]
        protocol.makeConnection(transport)

        clock.advance(120)
        self.assertEqual(results, [True])
        assert_clock_idle(self, clock)

    def test_disconnect_forwarded(self) -> None:
        """
        If the connection is closed, the underlying protocol is informed.
        """
        clock = Clock()
        wrapped_protocol = AccumulatingProtocol()
        protocol = HangCheckProtocol(wrapped_protocol, reactor=clock)
        transport = StringTransport()

        transport.protocol = protocol  # type: ignore[attr-defined]
        protocol.makeConnection(transport)
        reason = ConnectionDone("Bye.")
        protocol.connectionLost(Failure(reason))

        self.assertTrue(wrapped_protocol.closed)
        assert wrapped_protocol.closedReason is not None
        self.assertEqual(
            wrapped_protocol.closedReason.value,
            reason,
        )

    def test_disconnect_cancels_timeout(self) -> None:
        """
        If the connection is closed, the hang check is cancelled.
        """
        clock = Clock()
        protocol = HangCheckProtocol(
            Protocol(),
            reactor=clock,
        )
        transport = StringTransport()

        transport.protocol = protocol  # type: ignore[attr-defined]
        protocol.makeConnection(transport)
        protocol.connectionLost(Failure(ConnectionDone("Bye.")))

        assert_clock_idle(self, clock)

    def test_data_and_disconnect(self) -> None:
        """
        If the connection receives data and then is closed, no error results.
        """
        clock = Clock()
        protocol = HangCheckProtocol(
            Protocol(),
            reactor=clock,
        )
        transport = StringTransport()

        transport.protocol = protocol  # type: ignore[attr-defined]
        protocol.makeConnection(transport)
        protocol.dataReceived(b"some-data")
        protocol.connectionLost(Failure(ConnectionDone("Bye.")))

        assert_clock_idle(self, clock)


@defer.inlineCallbacks
def connected_server_and_client(
    case: TestCase,
    server_factory: SiteWithClose,
    client_factory: HangCheckFactory,
) -> InlineCallbacksType[None]:
    """
    Create a server and client connected to that server.

    :param case: The test case that will handle cleanup.
    :param IProtocolFactory server_factory: The factory for the server protocol.
    :param IProtocolFactory client_factory: The factory for the client protocol.

    :return: A deferred that fires when the client has connected.

    .. todo:

       Figure out what a sensible value to return is. The existing caller doesn't
       use the return value.
    """
    try:
        listening_port = yield TCP4ServerEndpoint(reactor, 0).listen(server_factory)
        case.addCleanup(listening_port.stopListening)

        endpoint = TCP4ClientEndpoint(reactor, '127.0.0.1', listening_port.getHost().port)
        yield endpoint.connect(client_factory)

    except Exception as e:
        f = Failure(e)  # we can't use `e` from the lambda itself
        case.addCleanup(lambda: f)


class EndToEndHangCheckTests(TestCase):
    # On slower machines with high CPU oversubscription this test may take longer to run than
    # the default timeout.
    timeout = 20

    @defer.inlineCallbacks
    def test_http(self) -> InlineCallbacksType[None]:
        """
        When connecting to a HTTP server, a PB connection times
        out.
        """
        result: defer.Deferred[None] = defer.Deferred()

        site = SiteWithClose(Data("", "text/plain"))
        client = HangCheckFactory(PBClientFactory(), lambda: result.callback(None))

        self.patch(HangCheckProtocol, '_HUNG_CONNECTION_TIMEOUT', 0.1)

        d_connected = connected_server_and_client(
            self,
            site,
            client,
        )

        def cancel_all() -> None:
            result.cancel()
            d_connected.cancel()

        timer = cast("IReactorTime", reactor).callLater(2, cancel_all)

        try:
            yield result
        except defer.CancelledError as e:
            raise RuntimeError('Timeout did not happen') from e
        finally:
            d_connected.cancel()
            timer.cancel()

        site.close_connections()
